from __future__ import print_function
import numpy as np
import sqaod
from sqaod.common.dense_graph_annealer_base import DenseGraphAnnealerBase
from sqaod.common import docstring
from . import cpu_dg_annealer as cext

class DenseGraphAnnealer(DenseGraphAnnealerBase) :
    """
    CPU-based dense graph annealer
    """

    def __init__(self, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        DenseGraphAnnealerBase.__init__(self, cext, dtype, W, optimize, prefdict)
        

def dense_graph_annealer(W = None, optimize=sqaod.minimize, dtype=np.float64, **prefs) :
    """ factory function for sqaod.cpu.DenseGraphAnnealer_.

    Args:
      numpy.ndarray W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs

    Returns:
      sqaod.cpu.DenseGraphAnnealer_: annealer instance

    """
    return DenseGraphAnnealer(W, optimize, dtype, prefs)


# refer docstring from interface
docstring.inherit(DenseGraphAnnealer, sqaod.py.DenseGraphAnnealer)


if __name__ == '__main__' :

    W = np.array([[-32,4,4,4,4,4,4,4],
                  [4,-32,4,4,4,4,4,4],
                  [4,4,-32,4,4,4,4,4],
                  [4,4,4,-32,4,4,4,4],
                  [4,4,4,4,-32,4,4,4],
                  [4,4,4,4,4,-32,4,4],
                  [4,4,4,4,4,4,-32,4],
                  [4,4,4,4,4,4,4,-32]])
    np.random.seed(0)

    N = 8
    m = 4

#    N = 10
#    m = 5
#    W = sqaod.generate_random_symmetric_W(N, -0.5, 0.5, np.float64)

    ann = dense_graph_annealer(W, dtype=np.float64, n_trotters = m)
    #import sqaod.py as py
    #ann = py.dense_graph_annealer(W, n_trotters = m)
    h, J, c = ann.get_hamiltonian()
    print(h)
    print(J)
    print(c)
    
    Ginit = 5.
    Gfin = 0.001
    
    nRepeat = 4
    beta = 1. / 0.02
    tau = 0.995
    
    for loop in range(0, nRepeat) :
        G = Ginit
        ann.set_preferences(n_trotters = 4)
        ann.prepare()
        ann.randomize_spin()
        while Gfin < G :
            ann.anneal_one_step(G, beta)
            G = G * tau
        ann.make_solution()
        E = ann.get_E()
        q = ann.get_q() 
        print(E)
        print(q)
        print(ann.get_preferences())



# operations to switch minimize / maximize.

class Algorithm :
    default = 'default'
    naive = 'naive'
    coloring = 'coloring'
    brute_force_search = 'brute_force_search'
    sa_naive = 'sa_naive'
    sa_coloring = 'sa_coloring'
    
algorithm = Algorithm()
"""
  sqaod.common.Algorithm: signatures to select algorithms

  algorithm signatures are provided as attributes shown below:

  **default:**
    select default algorithm.
    Each solver class has its default algorithm.  By specifying algorithm.default, 
    the default algirhtm will be chosen,
  **naive :**
    Naive version of SQA algorithm,
  **coloring :**
    SQA algorithm is parallelized by using color-coding,
  **brute_force_search :**
    Brute-force search.
    All brute-force searchers have 'brute_force_search' algorithm,
  **sa_naive :**
    naive version of SA(simulated annealing, not SQA),
  **sa_coloring :**
    color-code version of SA.

"""



class Minimize :
    """Tag for optimize direction.  This class also has some methods for sign manipulations."""
    @staticmethod
    def sign(v) :
        return v.copy()
    @staticmethod
    def best(list) :
        return min(list)
    @staticmethod
    def sort(list) :
        return sorted(list)
    def __int__(self) :
        return 0

    
class Maximize :
    """Tag for optimize direction.  This class also has some methods for sign manipulations."""
    @staticmethod
    def sign(v) :
        return -v
    @staticmethod
    def best(list) :
        return max(list)
    @staticmethod
    def sort(list) :
        return sorted(list, reverse = true)
    def __int__(self) :
        return 1

minimize = Minimize()  #: telling solvers to minimize QUBO energy.
maximize = Maximize()  #: telling solvers to maximize QUBO energy.



from __future__ import print_function

__version__ = '1.0.1'

from .common.envcheck import *
from .common.cuda_probe import is_cuda_available

def run_env_check() :
    print('CUDA : {}'.format('Available' if is_cuda_available() else 'Not available'))

    checker = EnvChecker(__version__)
    checker.check()
    checker.show()

try :
    # imports
    from .common.preference import algorithm
    from .common.preference import minimize
    from .common.preference import maximize

    from .common import *
    from . import py
    from . import cpu

    if is_cuda_available() :
        from . import cuda
    
except:
    checker = EnvChecker(__version__)
    if not checker.check() :
        print('\nWrong installation of sqaod backend libraries.')
        print('Please visit https://github.com/shinmorino/sqaod/wiki/installation to check installation instruction.\n')
        checker.show()



from __future__ import print_function
from __future__ import division
import numpy as np
import sqaod
from . import formulas
from sqaod.common import checkers, symmetrize
from types import MethodType
from sqaod import algorithm as algo

class DenseGraphAnnealer :
    """ python implementation of dense graph annealer.

    This class is a reference implementation of dense graph annealers.
    All algorithms are written in python.
    """
    def __init__(self, W, optimize, prefdict) :
        if not W is None :
            self.set_qubo(W, optimize)
        self._select_algorithm(algo.coloring)
        self.set_preferences(prefdict)

    def seed(self, seed) :
        """ set a seed value for random number generators.

        Args:
          seed (int or long) :
            seed value

        Note:
          sqaod.py annealer uses global random number genrator, so seed value is ignored.
          For annealers in other packages, given seed value is used as documented.  

        """
        # py version uses using global random generator.
        # Nothing to do here.
        pass
        
    def _vars(self) :
        return self._h, self._J, self._c, self._q
    
    def get_problem_size(self) :
        """ get problem size.

        Problem size is defined as a number of bits of QUBO.

        Returns:
          int: problem size.
        """
        return self._N;

    def set_qubo(self, W, optimize = sqaod.minimize) :
        """ set QUBO.

        Args:
          numpy.ndarray W :
            QUBO matrix.  W should be a sqaure matrix.
            Upper/Lower triangular matrices or symmetric matrices are accepted.
          optimize : optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`        """
        
        checkers.dense_graph.qubo(W)
        W = symmetrize(W)

        h, J, c = formulas.dense_graph_calculate_hamiltonian(W)
        self._optimize = optimize
        self._h, self._J, self._c = optimize.sign(h), optimize.sign(J), optimize.sign(c)
        self._N = W.shape[0]
        self._m = self._N // 2
        
    def set_hamiltonian(self, h, J, c) :
        """ set Hamiltonian.

        Args:
          numpy.ndarray : h(vector), J(matrix)
          floating point number : c

          Shapes for h, J must be (N), (N, N) where N is problem size.

          J must be a upper/lower triangular or symmetric matrix.
        """
        
        checkers.dense_graph.hJc(h, J, c)
        J = symmetrize(J)
        self._optimize = sqaod.minimize
        self._h, self._J, self._c = h, J, c
        self._N = J.shape[0]
        self._m = self._N // 2
        
    def _select_algorithm(self, algoname) :
        if algoname == algo.coloring :
            self.anneal_one_step = \
                MethodType(DenseGraphAnnealer.anneal_one_step_coloring, self)
        elif algoname == algo.sa_naive :
            self.anneal_one_step = \
                MethodType(DenseGraphAnnealer.anneal_one_step_sa_naive, self)
        else :
            self.anneal_one_step = \
                MethodType(DenseGraphAnnealer.anneal_one_step_naive, self)

    def _get_algorithm(self) :
        if self.anneal_one_step.__func__ == DenseGraphAnnealer.anneal_one_step_coloring :
            return algo.coloring;
        if self.anneal_one_step.__func__ == DenseGraphAnnealer.anneal_one_step_sa_naive :
            return algo.sa_naive;
        return algo.naive
            
    def set_preferences(self, prefdict = None, **prefs) :
        """ set solver preferences.

        Args:
          prefdict(dict) : preference dictionary.
          prefs(dict) : preference dictionary as \*\*kwargs. 

        References:
          `preference <preference.html>`_
        """
        if not prefdict is None :
            self._set_prefdict(prefdict)
        self._set_prefdict(prefs)
        
    def _set_prefdict(self, prefdict) :
        v = prefdict.get('n_trotters')
        if v is not None :
            self._m = v;
        v = prefdict.get('algorithm')
        if v is not None :
            self._select_algorithm(v)

    def get_preferences(self) :
        """ get solver preferences.

        Returns:
          dict: preference dictionary.

        References:
          `preference <preference.html>`_
        """
        prefs = { }
        if hasattr(self, '_m') :
            prefs['n_trotters'] = self._m
        prefs['algorithm'] = self._get_algorithm()
        return prefs
        
    def get_optimize_dir(self) :
        """ get optimize direction
        
        Returns:
          optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.

        """
        return self._optimize

    def get_E(self) :
        """ get QUBO energy.

        Returns:
          array of floating point number : QUBO energy.

        QUBO energy value is calculated for each trotter. so E is a vector whose length is m.

        Notes:
          You need to call calculate_E() or make_solution() before calling get_E().
          CPU/CUDA versions of solvers automatically call calculate_E() in get_E()
        """
        return self._E

    def get_x(self) :
        """ get bits.

        Returns:
          numpy.int8 : array of bit {0, 1}.

        x.shape is (m, N) where N is problem size, and m is number of trotters.

        Note:
          calculate_E() or make_solution() should be called before calling get_E().
          ( CPU/CUDA annealers automatically/internally call calculate_E().)
        """
        return self._x

    def set_q(self, q) :
        """ set spins.

        Args:
          q(np.array of np.int8) : array (length = N) of spin {-1, 1}.
        """

        if q.dtype != np.int8 :
            q = np.asarray(q, np.int8)
        self._q[:] = q

    def set_qset(self, q) :
        """ set "array of" spins.
        
        Args:
          qset (list of np.int8 array ) : list of array of spin {-1, 1}.


        qset is a list (length = m) of int8 array (length = N).
        """
        
        self._m = len(q)
        self.prepare()
        qlist = q
        for idx in range(len(qlist)) :
            q = qlist[idx]
            if q.dtype != np.int8 :
                q = np.asarray(q, np.int8)
            self._q[idx] = q
    
    # Ising model

    def get_hamiltonian(self) :
        """ get hamiltonian.
        
        Returns: tuple of Hamiltonian variables
          h(vector), J(symmeric matrix), c(scalar)
        """
        return self._h, self._J, self._c

    def get_q(self) :
        """ get spins.
        
        Returns:
          numpy.int8 : array of spin {-1, 1}.
        """
        return self._q
    
    def randomize_spin(self) :
        """ randomize spin. """
        sqaod.randomize_spin(self._q)

    def calculate_E(self) :
        """ calculate QUBO energy.
        
        This method calculate QUBO energy, and caches it, does not return any value.

        Note:
          A call to this method can be asynchronous.
        """
        
        h, J, c, q = self._vars()
        E = formulas.dense_graph_batch_calculate_E_from_spin(h, J, c, q)
        self._E = self._optimize.sign(E)

    def prepare(self) :
        """ preparation of internal resources.
        
        Note:
          prepare() should be called prior to run annealing loop.
        """
        if self._m == 1 :
            self._select_algorithm(algo.sa_naive)
        self._q = np.empty((self._m, self._N), dtype=np.int8)

    def make_solution(self) :
        """ make bit arrays(x) and calculate QUBO energy.

        Note:
          A call to this method can be asynchronous.
        """
        self._x = []
        for idx in range(self._m) :
            x = sqaod.bit_from_spin(self._q[idx])
            self._x.append(x)
        self.calculate_E()

    def anneal_one_step(self, G, beta) :
        """ Run annealing one step.

        Args:
          G (floating point number) : G in SQA, kT in SA.
          beta (floating point number) : inverse temperature.
        """

        # will be dynamically replaced.
        pass
        
    def anneal_one_step_naive(self, G, beta) :
        """ (sqaod.py only) sqaod.algorithm.naive version of SQA """
        
        h, J, c, q = self._vars()
        N = self._N
        m = self._m
        two_div_m = 2. / np.float64(m)
        coef = np.log(np.tanh(G * beta / m)) * beta
        
        for i in range(self._N * self._m):
            x = np.random.randint(N)
            y = np.random.randint(m)
            qyx = q[y][x]
            sum = np.dot(J[x], q[y]); # diagnoal elements in J are zero.
            dE = two_div_m * qyx * (h[x] + sum)
            dE -= qyx * (q[(m + y - 1) % m][x] + q[(y + 1) % m][x]) * coef
            threshold = 1. if (dE <= 0.) else np.exp(-dE * beta)
            if threshold > np.random.rand():
                q[y][x] = - qyx

    def anneal_colored_plane(self, G, beta, offset) :
        """ (sqaod.py only) try to flip spins in a colored plane. """
        h, J, c, q = self._vars()
        N = self._N
        m = self._m
        two_div_m = 2. / np.float64(m)
        coef = np.log(np.tanh(G * beta / m)) * beta
        
        for y in range(self._m):
            x = (offset + np.random.randint(1 << 30) * 2) % N
            qyx = q[y][x]
            sum = np.dot(J[x], q[y]); # diagnoal elements in J are zero.
            dE = two_div_m * qyx * (h[x] + sum)
            dE -= qyx * (q[(m + y - 1) % m][x] + q[(y + 1) % m][x]) * coef
            threshold = 1. if (dE <= 0.) else np.exp(-dE * beta)
            if threshold > np.random.rand():
                q[y][x] = - qyx
            
    def anneal_one_step_coloring(self, G, beta) :
        """ (sqaod.py only) sqaod.algorithm.coloring version of SQA """
        for loop in range(0, self._N) :
            self.anneal_colored_plane(G, beta, 0)
            self.anneal_colored_plane(G, beta, 1)
    
    def anneal_one_step_sa_naive(self, kT, beta) :
        """ (sqaod.py only) sqaod.algorithm.sa_naive version of SA """
        h, J, c, q = self._vars()
        N = self._N

        for iq in range(self._m) :
            qm = q[iq]
            for i in range(self._N):
                x = np.random.randint(N)
                qx = qm[x]
                sum = np.dot(J[x], qm); # diagnoal elements in J are zero.
                dE = 2. * qx * (h[x] + sum)
                threshold = 1. if (dE <= 0.) else np.exp(-dE * kT * beta)
                if threshold > np.random.rand():
                    qm[x] = - qx

                
def dense_graph_annealer(W = None, optimize = sqaod.minimize, **prefs) :
    """ factory function for sqaod.py.DenseGraphAnnealer.

    Args:
      numpy.ndarray : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.py.DenseGraphAnnealer: annealer instance
    """
    an = DenseGraphAnnealer(W, optimize, prefs)
    return an


if __name__ == '__main__' :

    np.random.seed(0)
    Ginit = 5.
    Gfin = 0.01
    
    nRepeat = 4
    beta = 1. / 0.02
    tau = 0.99
    
    N = 8
    m = 4
    W = np.array([[-32,4,4,4,4,4,4,4],
                  [4,-32,4,4,4,4,4,4],
                  [4,4,-32,4,4,4,4,4],
                  [4,4,4,-32,4,4,4,4],
                  [4,4,4,4,-32,4,4,4],
                  [4,4,4,4,4,-32,4,4],
                  [4,4,4,4,4,4,-32,4],
                  [4,4,4,4,4,4,4,-32]])
    

    #N = 20
    #m = 10
    #W = sqaod.generate_random_symmetric_W(N, -0.5, 0.5, np.float64)
    
    algoname = algo.default
    # algo = DenseGraphAnnealer.naive
    ann = dense_graph_annealer(W, sqaod.minimize, n_trotters=m)
    ann.set_preferences(algorithm = algo.naive)
    
    for loop in range(0, nRepeat) :
        G = Ginit
        ann.prepare()
        ann.randomize_spin()
        while Gfin < G :
            ann.anneal_one_step(G, beta)
            G = G * tau

        ann.make_solution()
        E = ann.get_E()
        #x = ann.get_x()
        print(E) # ,x

    prefs = ann.get_preferences()
    print(prefs)
        
    ann = dense_graph_annealer(W, sqaod.maximize)
    ann.set_preferences(prefs)
    for loop in range(0, nRepeat) :
        G = Ginit
        ann.prepare()
        ann.randomize_spin()
        while Gfin < G :
            ann.anneal_one_step(G, beta)
            G = G * tau

        ann.make_solution()
        E = ann.get_E()
        #x = ann.get_x()
        print(E) # ,x

    prefs = ann.get_preferences()
    print(prefs)



from __future__ import print_function
from __future__ import division
import numpy as np
from types import MethodType
import sqaod
from sqaod import algorithm as algo
from sqaod.common import checkers
from . import formulas


class BipartiteGraphAnnealer :
    """ python implementation of bipartite graph annealer.

    This class is a reference implementation of bipartite graph annealers.
    All algorithms are written in python.
    """

    def __init__(self, b0, b1, W, optimize, prefdict) : # n_trotters
        if not W is None :
            self.set_qubo(b0, b1, W, optimize)
        self._select_algorithm(algo.coloring)    
        self.set_preferences(prefdict)

    def _vars(self) :
        return self._h0, self._h1, self._J, self._c, self._q0, self._q1

    def seed(self, seed) :
        """ set a seed value for random number generators.

        Args:
          seed (int or long) :
            seed value

        Note:
          sqaod.py annealer uses global random number genrator, so seed value is ignored.
          For annealers in other packages, given seed value is used as documented.  

        """
        # this annealer use global random number genertor.
        # nothing to do here.
        pass
    
    def get_problem_size(self) :
        """ get problem size.

        Problem size is defined as a number of bits of QUBO.

        Returns:
          tuple containing problem size, (N0, N1).
        """
        return self._N0, self._N1;
        
    def set_qubo(self, b0, b1, W, optimize = sqaod.minimize) :
        """ set QUBO.

        Args:
          numpy.ndarray b0, b1, W : QUBO.
          optimize : optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
        """
        checkers.bipartite_graph.qubo(b0, b1, W)

        self._N0 = W.shape[1]
        self._N1 = W.shape[0]
        self._m = (self._N0 + self._N1) // 4
        self._optimize = optimize
        h0, h1, J, c = formulas.bipartite_graph_calculate_hamiltonian(b0, b1, W)
        self._h0, self._h1 = optimize.sign(h0), optimize.sign(h1)
        self._J, self._c = optimize.sign(J), optimize.sign(c)

    def set_hamiltonian(self, h0, h1,J, c) :
        """ set Hamiltonian.

        Args:
          numpy.ndarray : h0(vector), h1(vector), J(matrix)
          floating point number : c

          Shapes for h0, h1, J must be (N0), (N1), (N1, N0) where N0 and N1 are number of bits.
        """
        checkers.bipartite_graph.hJc(h0, h1, J, c)

        self._N0 = J.shape[1]
        self._N1 = J.shape[0]
        self._m = (self._N0 + self._N1) // 4
        self._optimize = sqaod.minimize
        self._h0, self._h1, self._J, self._c = h0, h1, J, c
        
    def _select_algorithm(self, algoname) :
        if algoname == algo.coloring :
            self.anneal_one_step = \
                MethodType(BipartiteGraphAnnealer.anneal_one_step_coloring, self)
        elif algoname == algo.sa_naive :
            self.anneal_one_step = \
                MethodType(BipartiteGraphAnnealer.anneal_one_step_sa_naive, self)
        elif algoname == algo.sa_coloring :
            self.anneal_one_step = \
                MethodType(BipartiteGraphAnnealer.anneal_one_step_sa_coloring, self)
        else : # algoname == algo.naive :
            self.anneal_one_step = \
                MethodType(BipartiteGraphAnnealer.anneal_one_step_naive, self)

    def _get_algorithm(self) :
        if self.anneal_one_step.__func__ == BipartiteGraphAnnealer.anneal_one_step_naive :
            return algo.naive;
        elif self.anneal_one_step.__func__ == BipartiteGraphAnnealer.anneal_one_step_sa_naive :
            return algo.sa_naive;
        elif self.anneal_one_step.__func__ == BipartiteGraphAnnealer.anneal_one_step_sa_coloring :
            return algo.sa_coloring;
        return algo.coloring
            
    def set_preferences(self, prefdict = None, **prefs) :
        """ set solver preferences.

        Args:
          prefdict(dict) : preference dictionary.
          prefs(dict) : preference dictionary as \*\*kwargs. 

        References:
          `preference <preference.html>`_
        """
        if not prefdict is None :
            self._set_prefdict(prefdict)
        self._set_prefdict(prefs)
        
    def _set_prefdict(self, prefdict) :
        v = prefdict.get('n_trotters')
        if v is not None :
            self._m = v;
        v = prefdict.get('algorithm')
        if v is not None :
            self._select_algorithm(v)

    def get_preferences(self) :
        """ get solver preferences.

        Returns:
          dict: preference dictionary.

        References:
          `preference <preference.html>`_
        """
        prefs = { }
        if hasattr(self, '_m') :
            prefs['n_trotters'] = self._m
        prefs['algorithm'] = self._get_algorithm()
        return prefs

    def get_optimize_dir(self) :
        """ get optimize direction
        
        Returns:
          optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
        """
        return self._optimize
        
    def get_E(self) :
        """ get QUBO energy.

        Returns:
          array of floating point number : QUBO energy.

        QUBO energy value is calculated for each trotter. so E is a vector whose length is m.

        Notes:
          You need to call calculate_E() or make_solution() before calling get_E().
          CPU/CUDA versions of solvers automatically call calculate_E() in get_E()
        """
        return self._E
    
    def get_x(self) :
        """ get bits.

        Returns:
          tuple of 2 numpy.int8 arrays : array of bit {0, 1}.

        x0.shape and x1.shape are (m, N0) and (m, N1) repsectively, and m is number of trotters.

        Note:
          calculate_E() or make_solution() should be called before calling get_E().
          ( CPU/CUDA annealers automatically/internally call calculate_E().)
        """
        return self._x_pairs

    def set_q(self, qpair) :
        self.prepare()
        for idx in range(0, self._m) :
            self._q0[idx] = qpair[0]
            self._q1[idx] = qpair[1]

    def set_qset(self, qpair) :
        qpairs = qpair
        self._m = len(qpairs);
        self.prepare()
        for idx in range(0, self._m) :
            self._q0[idx] = qpairs[idx][0]
            self._q1[idx] = qpairs[idx][1]

    # Ising model / spin
    
    def get_hamiltonian(self) :
        """ get hamiltonian.
        
        Returns: tuple of Hamiltonian variables
          h0(vector), h1(vector), J(matrix), c(scalar)
        """
        return self._h0, self._h1, self._J, self._c
            
    def get_q(self) :
        """ get spins.
        
        Returns: tuple for (q0, q1).  The q0 and q1 are numpy.int8 arrays whose lengths are N0 and N1 respectively.
        """
        qlist = []
        for qpair in zip(self._q0, self._q1) :
            qlist.append(qpair)
        return qlist

    def randomize_spin(self) :
        """ randomize spin. """
        sqaod.randomize_spin(self._q0)
        sqaod.randomize_spin(self._q1)

    def calculate_E(self) :
        """ calculate QUBO energy.
        
        This method calculate QUBO energy, and caches it, does not return any value.

        Note:
          A call to this method can be asynchronous.
        """
        h0, h1, J, c, q0, q1 = self._vars()
        E = np.empty((self._m), J.dtype)
        for idx in range(self._m) :
            # FIXME: 1d output for batch calculation
            E[idx] = formulas.bipartite_graph_calculate_E_from_spin(h0, h1, J, c, q0[idx], q1[idx])
        self._E = self._optimize.sign(E)

    def prepare(self) :
        """ preparation of internal resources.
        
        Note:
          prepare() should be called prior to run annealing loop.
        """
        self._q0 = np.empty((self._m, self._N0), dtype=np.int8)
        self._q1 = np.empty((self._m, self._N1), dtype=np.int8)
        if self._m == 1 :
            cur_algo = self._get_algorithm()
            if  cur_algo != algo.sa_naive and cur_algo != algo.sa_coloring :
                self._select_algorithm(algo.sa_naive)

    def make_solution(self) :
        """ make bit arrays(x) and calculate QUBO energy.

        Note:
          A call to this method can be asynchronous.
        """
        self._x_pairs = []
        for idx in range(self._m) :
            x0 = sqaod.bit_from_spin(self._q0[idx])
            x1 = sqaod.bit_from_spin(self._q1[idx])
            self._x_pairs.append((x0, x1))
        self.calculate_E()

    def anneal_one_step(self, G, beta) :
        """ Run annealing one step.

        Args:
          G (floating point number) : G in SQA, kT in SA.
          beta (floating point number) : inverse temperature.
        """
        # will be dynamically replaced.
        pass
                
    def anneal_one_step_naive(self, G, beta) :
        """ (sqaod.py only) sqaod.algorithm.naive version of SQA """
        h0, h1, J, c, q0, q1 = self._vars()
        N0, N1 = self.get_problem_size()
        m = self._m
        N = N0 + N1
        twoDivM = 2. / m
        tempCoef = np.log(np.tanh(G * beta / m)) * beta

        for loop in range(N * m) :
            iq = np.random.randint(N)
            im = np.random.randint(m)
            mNeibour0 = (im + m - 1) % m
            mNeibour1 = (im + 1) % m
            
            if (iq < N0) :
                q = q0[im][iq]
                dE = twoDivM * q * (h0[iq] + np.dot(J.T[iq], q1[im]))
                dE -= q * (q0[mNeibour0][iq] + q0[mNeibour1][iq]) * tempCoef
                thresh = 1 if dE < 0 else np.exp(-dE * beta) 
                if thresh > np.random.rand():
                    q0[im][iq] = -q
            else :
                iq -= N0
                q = q1[im][iq]
                dE = twoDivM * q * (h1[iq] + np.dot(J[iq], q0[im]))
                dE -= q * (q1[mNeibour0][iq] + q1[mNeibour1][iq]) * tempCoef
                thresh = 1 if dE < 0 else np.exp(-dE * beta) 
                if thresh > np.random.rand():
                    q1[im][iq] = -q
        
    def _anneal_half_step_coloring(self, N, qAnneal, h, J, qFixed, G, beta, m) :
        """ (sqaod.py only) try to flip spins in a colored plane for SQA. """
        dEmat = np.matmul(J, qFixed.T)
        twoDivM = 2. / m
        tempCoef = np.log(np.tanh(G * beta / m)) * beta
        for im in range(0, m, 2) :
            mNeibour0 = (im + m - 1) % m
            mNeibour1 = (im + 1) % m
            for iq in range(0, N) :
                q = qAnneal[im][iq]
                dE = twoDivM * q * (h[iq] + dEmat[iq, im])
                dE -= q * (qAnneal[mNeibour0][iq] + qAnneal[mNeibour1][iq]) * tempCoef
                thresh = 1 if dE < 0 else np.exp(-dE * beta) 
                if thresh > np.random.rand():
                    qAnneal[im][iq] = -q
        for im in range(1, m, 2) :
            mNeibour0 = (im + m - 1) % m
            mNeibour1 = (im + 1) % m
            for iq in range(0, N) :
                q = qAnneal[im][iq]
                dE = twoDivM * q * (h[iq] + dEmat[iq, im])
                dE -= q * (qAnneal[mNeibour0][iq] + qAnneal[mNeibour1][iq]) * tempCoef
                thresh = 1 if dE < 0 else np.exp(-dE * beta) 
                if thresh > np.random.rand():
                    qAnneal[im][iq] = -q
                
    def anneal_one_step_coloring(self, G, beta) :
        """ (sqaod.py only) try to flip spins in a colored plane. """
        h0, h1, J, c, q0, q1 = self._vars()
        N0, N1 = self.get_problem_size()
        m = self._m
        self._anneal_half_step_coloring(N1, q1, h1, J, q0, G, beta, m)
        self._anneal_half_step_coloring(N0, q0, h0, J.T, q1, G, beta, m)

    # simulated annealing
    def anneal_one_step_sa_naive(self, kT, beta) :
        """ (sqaod.py only) sqaod.algorithm.sa_naive version of SA """
        h0, h1, J, c, q0, q1 = self._vars()
        N0, N1 = self.get_problem_size()
        m = self._m
        N = N0 + N1

        for im in range(m) :
            q0m, q1m = q0[im],q1[im]

            for loop in range(N) :
                iq = np.random.randint(N)
                if (iq < N0) :
                    q = q0m[iq]
                    dE = 2. * q * (h0[iq] + np.dot(J.T[iq], q1m))
                    thresh = 1 if dE < 0 else np.exp(-dE * kT * beta) 
                    if thresh > np.random.rand():
                        q0m[iq] = -q
                else :
                    iq -= N0
                    q = q1m[iq]
                    dE = 2. * q * (h1[iq] + np.dot(J[iq], q0m))
                    thresh = 1 if dE < 0 else np.exp(-dE * kT * beta) 
                    if thresh > np.random.rand():
                        q1m[iq] = -q
                        
    def _anneal_half_step_sa_coloring(self, N, qAnneal, h, J, qFixed, kT, beta, m) :
        """ (sqaod.py only) try to flip spins in a colored plane for SA. """
        dEmat = np.matmul(J, qFixed.T)
        for im in range(m) :
            qAnnealm = qAnneal[im]
            for iq in range(0, N) :
                q = qAnnealm[iq]
                dE = 2. * q * (h[iq] + dEmat[iq, im])
                thresh = 1 if dE < 0 else np.exp(-dE * kT * beta) 
                if thresh > np.random.rand():
                    qAnnealm[iq] = -q
                
    def anneal_one_step_sa_coloring(self, kT, beta) :
        """ (sqaod.py only) sqaod.algorithm.sa_coloring version of SA """
        h0, h1, J, c, q0, q1 = self._vars()
        N0, N1 = self.get_problem_size()
        m = self._m
        self._anneal_half_step_sa_coloring(N1, q1, h1, J, q0, kT, beta, m)
        self._anneal_half_step_sa_coloring(N0, q0, h0, J.T, q1, kT, beta, m)
        


def bipartite_graph_annealer(b0 = None, b1 = None, W = None, \
                             optimize = sqaod.minimize, **prefs) :
    """ factory function for sqaod.py.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.py.BipartiteGraphAnnealer: annealer instance
    """
    return BipartiteGraphAnnealer(b0, b1, W, optimize, prefs)


if __name__ == '__main__' :
    N0 = 12
    N1 = 10
    m = 10
    
    np.random.seed(0)
            
    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5

    an = bipartite_graph_annealer(b0, b1, W, sqaod.minimize, n_trotters=m, algorithm=algo.coloring)
    
    Ginit = 5
    Gfin = 0.01
    beta = 1. / 0.02
    tau = 0.99
    n_repeat = 10

    for loop in range(0, n_repeat) :
        an.prepare()
        an.randomize_spin()
        G = Ginit
        while Gfin < G :
            an.anneal_one_step(G, beta)
            G = G * tau
        an.make_solution()
        
        E = an.get_E()
        x = an.get_x()
        print(E)
        #print(x)



from __future__ import print_function
import sys

def is_cuda_available() :
    try :
        # loading driver
        import sqaod.cuda
    except :
        return False
    else :
        return True

def cuda_failure_reason() :
    try :
        # try loading cuda module.
        import sqaod.cuda
    except RuntimeError as e :
        print(e)
    except ImportError as e :
        print(e)
    except :
        print("Unexpected error:", sys.exc_info())
    else :
        return true, "OK"
        
if __name__ == '__main__' :
    cuda_available = is_cuda_available()
    print('cuda available : ' + str(cuda_available))
    if cuda_available is False :
        cuda_failure_reason()



import sys
import numbers
import numpy as np

import sys
this = sys.modules[__name__]

this.rtol_fp64 = 1.0e-5
this.atol_fp64 = 1.0e-8
this.rtol_fp32 = 1.0e-4
this.atol_fp32 = 1.0e-5

# common

def is_scalar(v) :
    if isinstance(v, numbers.Number) :
        return True
    if isinstance(v, number.ndarray) :
        if len(v.shape) == 1 :
            return v.shape[0] == 1
        elif len(v.shape) == 2 :
            return v.shape == (1, 1)
    return False

def is_vector(v) :
    if not isinstance(v, np.ndarray) :
        return False
    if len(v.shape) == 1 :
        return True
    if len(v.shape) == 2 :
        if v.shape[1] == 1 :
            return True
    return False

def is_bits(v) :
    for bits in bitslist :
        if bits.dtype != np.int8 :
            return False
    return True

def is_symmetric(mat, rtol = this.rtol_fp64, atol = this.atol_fp64) :
    this = sys.modules[__name__]
    if mat.dtype == np.float64 :
        rtol, atol = this.rtol_fp64, this.atol_fp64
    else :
        rtol, atol = this.rtol_fp32, this.atol_fp32
    return np.allclose(mat, mat.T, rtol, atol)

def is_triangular(mat) :
    atol = this.rtol_fp64 if mat.dtype == np.float64 else this.rtol_fp32
    if not np.any(np.triu(mat, 1) > atol) :
        return True
    return not np.any(np.tril(mat, -1) > atol)

def symmetrize(mat) :
    if is_symmetric(mat) :
        return mat
    
    if is_triangular(mat) :
        d = np.diag(mat)
        mat = mat + mat.T
        N = mat.shape[0]
        mat[range(N), range(N)] -= d
        return mat
    
    raise RuntimeError("given matrix is not triangular nor symmetric.")
    

def generate_random_symmetric_W(N, wmin = -0.5, wmax = 0.5, dtype=np.float64) :
    W = np.zeros((N, N), dtype)
    for i in range(0, N) :
        for j in range(i, N) :
            W[i, j] = np.random.random()
    W = W + np.tril(np.ones((N, N)), -1) * W.T
    W = W * (wmax - wmin) + wmin
    return W


# clone number/buffer object as specified as dtype.

def clone_as_number(v, dtype) :
    if isinstance(v, numbers.Number) :
        return dtype.type(v)
    if isinstance(v, number.ndarray) :
        if len(v.shape) == 1 and v.shape[0] == 1 :
            return dtype.type(v[0])
        elif len(v.shape) == 2 and v.shape == (1, 1) :
            return dtype.type(v[0][0])
    raise_not_a_scalar('v', v)

def create_bitset_sequence(vals, nbits) :
    seqlen = len(vals)
    x = np.ndarray((seqlen, nbits), np.int8)
    iseq = 0
    for v in vals :
        for pos in range(nbits) :
            x[iseq][pos] = np.int8(v >> (nbits - 1 - pos) & 1)
        iseq += 1
    return x

def _fix_ndarray_type(obj, dtype) :
    if obj.dtype == dtype and obj.flags['C_CONTIGUOUS'] :
        return obj
    return np.asarray(obj, dtype=dtype, order='C')

def fix_type(obj, dtype) :
    if isinstance(obj, np.ndarray) :
        return _fix_ndarray_type(obj, dtype)
    try :
        objs = []
        for nobj in obj :
            if isinstance(nobj, np.ndarray) :
                nobj = _fix_ndarray_type(nobj, dtype)
            else :
                # try creating ndarray from given object.
                nobj = np.asarray(nobj, dtype=dtype, order='C')
            objs.append(nobj)
        return objs
    except TypeError as te :
        raise RuntimeError('Fix failed.')
    

def generate_random_bits(N) :
    bits = np.empty((N), np.int8)
    for bit in np.nditer(bits, [], [['readwrite']]) :
        bit[...] = np.random.choice([0,1])
    return bits

def randomize_spin(qmat) :
    for spinvec in np.nditer(qmat, [], [['readwrite']]) :
        spinvec[...] = np.random.choice([-1,1])

def bit_to_spin(x) :
    q = x * 2 - 1
    return q

def bit_from_spin(q) :
    x = ((q + 1) >> 1)
    return x



def anneal(annealer, Ginit = 5., Gfin = 0.01, beta = 1. / 0.02, tau = 0.99, n_repeat = 10, verbose = False) :
    Emin = sys.float_info.max
    q0 = []
    q1 = []
    
    for loop in range(0, n_repeat) :
        annealer.prepare()
        annealer.randomize_spin()
        G = Ginit
        while Gfin < G :
            annealer.anneal_one_step(G, beta)
            if verbose :
                E - annealer.calculate_E()
                print(E)
            G = G * tau

        annealer.make_solution()



import numpy as np
import sqaod
from sqaod.common import checkers, symmetrize


def dense_graph_calculate_hamiltonian(W, dtype = None) :
    """ Calculate hamiltonian from given QUBO.

    Args:
      numpy.ndarray W : QUBO, W should be a upper/lower triangular or symmetric matrix.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      tuple: tuple containing Hamiltonian.
        h(vector as numpy.array), J(2-D symmetric matrix as numpy.array), c(scalar value)
    """
 
    checkers.dense_graph.qubo(W)
    W = symmetrize(W)

    N = W.shape[0]
    h = np.ndarray((N), dtype=np.float64)
    W4 = - 0.25 * W

    for i in range(N) :
        h[i] = - 0.5 * np.sum(W[i])
    c = np.sum(W4) + np.sum(W4.diagonal())

    J = W4
    for i in range(0, N) :
        J[i][i] = 0

    return h, J, c


def dense_graph_calculate_E(W, x, dtype = None) :
    """ Calculate desne graph QUBO energy from bits.

    Args:
      numpy.array W: QUBO, W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x : array of bit {0, 1}.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    
    checkers.dense_graph.qubo(W)
    checkers.dense_graph.x(W, x)
    W = symmetrize(W)

    if len(x.shape) != 1 :
        if x.shape[0] != 1 :
            raise Exception('Wrong dimention of x')
        x = x.reshape(-1)
    return np.dot(x, np.matmul(W, x.T))

def dense_graph_batch_calculate_E(W, x, dtype = None) :
    """ Batched version of the function to calculate dense graph QUBO energy from bits.

    Args:
      numpy.ndarray W: QUBO, W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x:
        bit arrays repsesented as 2-D matrix(n_trotters x n_bits).
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64

    Returns:
      QUBO energy
    """
    
    checkers.dense_graph.qubo(W)
    checkers.dense_graph.xbatch(W, x)
    W = symmetrize(W)
    
    if len(x.shape) == 1:
        x = x.reshape(1, -1)
    return np.sum(x * np.matmul(W, x.T).T, 1)

def dense_graph_calculate_E_from_spin(h, J, c, q, dtype = None) :
    """ Calculate desne graph QUBO energy from spins.

    Args:
      (numpy.ndarray) h, J, c : Hamiltonian h(1-D vector), W(sqauare matrix), c(scalar).
        W must be a upper/lower triangular or symmetric matrix.
      q (numpy.ndarray) : array of spin {-1, 1}.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64

    Returns:
      floating point number : QUBO energy
    """
    checkers.dense_graph.hJc(h, J, c)
    checkers.dense_graph.q(J, q)
    J = symmetrize(J)
    
    return - c - np.dot(h, q) - np.dot(q, np.matmul(J, q.T))

def dense_graph_batch_calculate_E_from_spin(h, J, c, q, dtype = None) :
    """ Batched version of the function to calculate dense graph QUBO energy from spins.

    Args:
      numpy.ndarray h, J, c : Hamiltonian h(1-D vector), W(sqauare matrix), c(scalar).
        W must be a upper/lower triangular or symmetric matrix.
      numpy.ndarray q :
        bit arrays repsesented as 2-D matrix(n_trotters x n_bits).
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    checkers.dense_graph.hJc(h, J, c)
    checkers.dense_graph.qbatch(J, q)
    J = symmetrize(J)
    
    if len(q.shape) == 1:
        q = q.reshape(1, -1)
    return - c - np.matmul(h, q.T) - np.sum(q.T * np.matmul(J, q.T), 0)


# bibartite graph

def bipartite_graph_calculate_hamiltonian(b0, b1, W, dtype = None) :
    """ Calculate hamiltonian from given QUBO.

    Args:
      numpy.ndarray b0, b1, W : Bipartite graph QUBO.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64

    Returns:
      tuple: tuple containing Hamiltonian.
        h0(1-D numpy.array), h1(1-D numpy.array), J(matrix as numpy.array), c(scalr value).
    """
    checkers.bipartite_graph.qubo(b0, b1, W);
    N0 = W.shape[1]
    N1 = W.shape[0]
    
    c = - 0.25 * np.sum(W) - 0.5 * (np.sum(b0) + np.sum(b1))
    J = - 0.25 * W
    h0 = np.empty((N0), W.dtype)
    h1 = np.empty((N1), W.dtype)
    for i in range(N0) :
        h0[i] = (- 1. / 4.) * np.sum(W[:, i]) - 0.5 * b0[i]
    for j in range(N1) :
        h1[j] = (- 1. / 4.) * np.sum(W[j]) - 0.5 * b1[j]

    return h0, h1, J, c

def bipartite_graph_calculate_E(b0, b1, W, x0, x1, dtype = None) :
    """ Calculate bipartite graph QUBO energy from bits.

    Args:
      numpy.array b0, b1, W: QUBO.  b0(vector as numpy.array), b1(vector as numpy.array), W(matrix as numpy.array).  W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x0, x1 : array of bit {0, 1}.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.x(W, x0, x1)
    return np.dot(b0, x0) + np.dot(b1, x1) + np.dot(x1, np.matmul(W, x0))

def bipartite_graph_batch_calculate_E(b0, b1, W, x0, x1, dtype = None) :
    """ Batched version of the function to calculate dense graph QUBO energy from bits.

    Args:
      numpy.array b0, b1, W: QUBO.  b0(vector as numpy.array), b1(vector as numpy.array), W(matrix as numpy.array).  W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x0, x1:
        bit arrays repsesented as 2-D matrix.
        x0 shape is n_trotters x n_bits_0, and x1 shape is n_trotters x n_bits_1
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    if len(x0.shape) == 1 :
        x0 = x0.reshape(1, -1)
    if len(x1.shape) == 1 :
        x1 = x1.reshape(1, -1)
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.xbatch(W, x0, x1)
    
    nBatch0, nBatch1 = x0.shape[0], x1.shape[0]
    if nBatch0 != nBatch1 :
        raise Exception("Different batch dims between x0 and x1.")
    return np.matmul(b0, x0.T) + np.matmul(b1, x1.T) + np.sum(x1 * np.matmul(W, x0.T).T, 1)

def bipartite_graph_batch_calculate_E_2d(b0, b1, W, x0, x1, dtype = None) :
    """ 2D batched version of the function to calculate dense graph QUBO energy from bits.

    Args:
      numpy.array b0, b1, W: QUBO.  b0(vector as numpy.array), b1(vector as numpy.array), W(matrix as numpy.array).  W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x0, x1:
        bit arrays repsesented as 2-D matrix.
        x0 shape is n_trotters x n_bits_0, and x1 shape is n_trotters x n_bits_1
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy as a matrix, whose shape is (len(x1), (len(x0)).

    This function calculates QUBO energy for all combinations of elements x0 and x1, 
    used in bipartite graph brute-force searchers.

    """
    if len(x0.shape) == 1 :
        x0 = x0.reshape(1, -1)
    if len(x1.shape) == 1 :
        x1 = x1.reshape(1, -1)
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.xbatch(W, x0, x1)
    
    return np.matmul(b0.T, x0.T).reshape(1, -1) + np.matmul(b1.T, x1.T).reshape(-1, 1) \
        + np.matmul(x1, np.matmul(W, x0.T))

def bipartite_graph_calculate_E_from_spin(h0, h1, J, c, q0, q1, dtype = None) :
    """ Calculate bipartite graph QUBO energy from spins.

    Args:
      numpy.array b0, b1, W: QUBO.  b0(vector as numpy.array), b1(vector as numpy.array), W(matrix as numpy.array).  W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x0, x1 : array of bit {0, 1}.
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    checkers.bipartite_graph.hJc(h0, h1, J, c)
    checkers.bipartite_graph.q(J, q0, q1)
    
    return - np.dot(h0, q0) - np.dot(h1, q1) - np.dot(q1, np.matmul(J, q0)) - c

def bipartite_graph_batch_calculate_E_from_spin(h0, h1, J, c, q0, q1, dtype = None) :
    """ Batched version of the function to calculate dense graph QUBO energy from bits.

    Args:
      numpy.array b0, b1, W: QUBO. b0(vector as numpy.array), b1(vector as numpy.array), W(matrix as numpy.array).  W should be a upper/lower triangular or symmetric matrix.
      numpy.ndarray x0, x1:
        bit arrays repsesented as 2-D matrix.
        x0 shape is n_trotters x n_bits_0, and x1 shape is n_trotters x n_bits_1
      numpy.dtype dtype : Numerical precision, numpy.float32 or numpy.float64.

    Returns:
      QUBO energy
    """
    if len(q0.shape) == 1 :
        q0 = q0.reshape(1, -1)
    if len(q1.shape) == 1 :
        q1 = q1.reshape(1, -1)
    checkers.bipartite_graph.hJc(h0, h1, J, c)
    checkers.bipartite_graph.qbatch(J, q0, q1)
    
    nBatch0, nBatch1 = q0.shape[0], q1.shape[0]
    if nBatch0 != nBatch1 :
        raise Exception("Different batch dims between x0 and x1.")
    return - np.matmul(h0, q0.T) - np.matmul(h1, q1.T) - np.sum(q1.T * np.matmul(J, q0.T), 0) - c



from __future__ import print_function
import numpy as np
import numbers
import sqaod
from sqaod.common import checkers
import sqaod.common as common

# dense graph    

# QUBO energy functions

def dense_graph_calculate_E(cext, cobj, W, x, dtype) :
    W = common.fix_type(W, dtype)
    W = common.symmetrize(W)
    x = common.fix_type(x, np.int8)

    checkers.dense_graph.qubo(W)
    checkers.dense_graph.x(W, x)
    
    E = np.ndarray((1), dtype)
    cext.dense_graph_calculate_E(cobj, E, W, x, dtype)
    return E[0]

def dense_graph_batch_calculate_E(cext, cobj, W, x, dtype) :
    W = common.fix_type(W, dtype)
    W = common.symmetrize(W)
    x = common.fix_type(x, np.int8)
    if len(x.shape) == 1 :
        x = x.reshape(1, -1)
    checkers.dense_graph.qubo(W)
    checkers.dense_graph.xbatch(W, x)
    
    E = np.empty((x.shape[0]), dtype)
    cext.dense_graph_batch_calculate_E(cobj, E, W, x, dtype)
    return E


# QUBO -> Ising model

def dense_graph_calculate_hamiltonian(cext, cobj, W, dtype) :
    W = common.fix_type(W, dtype)
    W = common.symmetrize(W)
    checkers.dense_graph.qubo(W)

    N = W.shape[0]
    h = np.empty((N), dtype)
    J = np.empty((N, N), dtype)
    c = np.empty((1), dtype)
    cext.dense_graph_calculate_hamiltonian(cobj, h, J, c, W, dtype);
    return h, J, c[0]

# Ising model energy functions

def dense_graph_calculate_E_from_spin(cext, cobj, h, J, c, q, dtype) :
    h, J = common.fix_type([h, J], dtype)
    J = common.symmetrize(J)
    q = common.fix_type(q, np.int8)
    checkers.dense_graph.hJc(h, J, c);
    checkers.dense_graph.q(J, q);
    
    E = np.ndarray((1), dtype)
    cext.dense_graph_calculate_E_from_spin(cobj, E, h, J, c, q, dtype)
    return E[0]

def dense_graph_batch_calculate_E_from_spin(cext, cobj, h, J, c, q, dtype) :
    h, J = common.fix_type([h, J], dtype)
    J = common.symmetrize(J)
    if len(q.shape) == 1 :
        q = q.reshape(1, -1)
    checkers.dense_graph.hJc(h, J, c);
    checkers.dense_graph.qbatch(J, q);

    E = np.empty([q.shape[0]], dtype)
    cext.dense_graph_batch_calculate_E_from_spin(cobj, E, h, J, c, q, dtype)
    return E


# bipartite_graph

def bipartite_graph_calculate_E(cext, cobj, b0, b1, W, x0, x1, dtype) :
    b0, b1, W = common.fix_type([b0, b1, W], dtype)
    x0, x1 = common.fix_type([x0, x1], np.int8)
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.x(W, x0, x1)
    
    E = np.ndarray((1), dtype)
    cext.bipartite_graph_calculate_E(cobj, E, b0, b1, W, x0, x1, dtype)
    return E[0]

def bipartite_graph_batch_calculate_E(cext, cobj, b0, b1, W, x0, x1, dtype) :
    b0, b1, W = common.fix_type([b0, b1, W], dtype)
    if len(x0.shape) == 1 :
        x0 = x0.reshape(1, -1)
    if len(x1.shape) == 1 :
        x1 = x1.reshape(1, -1)
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.xbatch(W, x0, x1)
    nBatch0, nBatch1 = x0.shape[0], x1.shape[0]
    if nBatch0 != nBatch1 :
        raise Exception("Different batch dims between x0 and x1.")
    
    E = np.empty((nBatch0), dtype)
    cext.bipartite_graph_batch_calculate_E(cobj, E, b0, b1, W, x0, x1, dtype)
    return E

def bipartite_graph_batch_calculate_E_2d(cext, cobj, b0, b1, W, x0, x1, dtype) :
    b0, b1, W = common.fix_type([b0, b1, W], dtype)
    x0, x1 = common.fix_type([x0, x1], np.int8)
    if len(x0.shape) == 1 :
        x0 = x0.reshape(1, -1)
    if len(x1.shape) == 1 :
        x1 = x1.reshape(1, -1)    
    checkers.bipartite_graph.qubo(b0, b1, W)
    checkers.bipartite_graph.xbatch(W, x0, x1)
    nBatch0, nBatch1 = q0.shape[0], q1.shape[0]
    
    E = np.empty((nBatch1, nBatch0), dtype)
    cext.bipartite_graph_batch_calculate_E_2d(cobj, E, b0, b1, W, x0, x1, dtype)
    return E


def bipartite_graph_calculate_hamiltonian(cext, cobj, b0, b1, W, dtype) :
    b0, b1, W = common.fix_type([b0, b1, W], dtype)
    checkers.bipartite_graph.qubo(b0, b1, W)

    N0 = W.shape[1]
    N1 = W.shape[0]
    h0 = np.empty((N0), dtype)
    h1 = np.empty((N1), dtype)
    J = np.empty((N1, N0), dtype)
    c = np.empty((1), dtype)
    cext.bipartite_graph_calculate_hamiltonian(cobj, h0, h1, J, c, b0, b1, W, dtype);
    return h0, h1, J, c[0]

def bipartite_graph_calculate_E_from_spin(cext, cobj, h0, h1, J, c, q0, q1, dtype) :
    h0, h1, W = common.fix_type([h0, h1, J], dtype)
    q0, q1 = common.fix_type([q0, q1], np.int8)
    checkers.bipartite_graph.hJc(h0, h1, J, c)
    checkers.bipartite_graph.q(J, q0, q1)

    E = np.ndarray((1), dtype)
    cext.bipartite_graph_calculate_E_from_spin(cobj, E, h0, h1, J, c, q0, q1, dtype)
    return E[0]

def bipartite_graph_batch_calculate_E_from_spin(cext, cobj, h0, h1, J, c, q0, q1, dtype) :
    h0, h1, W = common.fix_type([h0, h1, J], dtype)
    q0, q1 = common.fix_type([q0, q1], np.int8)
    if len(q0.shape) == 1 :
        q0 = q0.reshape(1, -1)
    if len(q1.shape) == 1 :
        q1 = q1.reshape(1, -1)    
    checkers.bipartite_graph.hJc(h0, h1, J, c)
    checkers.bipartite_graph.qbatch(J, q0, q1)
    nBatch0, nBatch1 = q0.shape[0], q1.shape[0]
    if nBatch0 != nBatch1 :
        raise Exception("Different batch dims between x0 and x1.")
        
    E = np.empty((nBatch0), dtype)
    cext.bipartite_graph_batch_calculate_E_from_spin(cobj, E, h0, h1, J, c, q0, q1, dtype)
    return E



from . import cuda_device
import sys

class Device :

    _cext = cuda_device

    def __init__(self, devno) :
        self._cobj = cuda_device.new()
        cuda_device.initialize(self._cobj, devno)

    def __del__(self) :
        if not self._cobj is None :
            self.finalize();
        self._cobj = None

    def finalize(self) :
        cuda_device.finalize(self._cobj);
        cuda_device.delete(self._cobj)
        self._cobj = None
        
def create_device(devno = 0) :
    return Device(devno)


# Global/Static

this_module = sys.modules[__name__]

def unload() :
    if not this_module.active_device is None :
        this_module.active_device = None

if __name__ != "__main__" :
    this_module.active_device = create_device()
    import atexit
    atexit.register(unload)



from __future__ import print_function
import numpy as np
import sys
import sqaod
from sqaod import algorithm as algo
from . import formulas
from sqaod.common import checkers, symmetrize

class DenseGraphBFSearcher :
    """ Dense graph brute-force searcher"""

    def __init__(self, W, optimize, prefdict) :
        if W is not None :
            self.set_qubo(W, optimize)
        self._tile_size = 1024
        self.set_preferences(prefdict)
            
    def set_qubo(self, W, optimize = sqaod.minimize) :
        """ set QUBO.

        Args:
          numpy.ndarray W :
            QUBO matrix.  W should be a sqaure matrix.
            Upper/Lower triangular matrices or symmetric matrices are accepted.
          optimize : optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`        """
        checkers.dense_graph.qubo(W)
        W = symmetrize(W)
        N = W.shape[0]
        self._W = optimize.sign(W)
        self._N = N
        self._x = []
        self._optimize = optimize

    def _get_algorithm(self) :
        return algo.brute_force_search
            
    def set_preferences(self, prefdict=None, **prefs) :
        """ set solver preferences.

        Args:
          prefdict(dict) : preference dictionary.
          prefs(dict) : preference dictionary as \*\*kwargs. 

        References:
          `preference <preference.html>`_
        """
        if not prefdict is None :
            self._set_prefdict(prefdict)
        self._set_prefdict(prefs)
        
    def _set_prefdict(self, prefdict) :
        v = prefdict.get('tile_size')
        if v is not None :
            self._tile_size = v;
    
    def get_problem_size(self) :
        """ get problem size.

        Problem size is defined as a number of bits of QUBO.

        Returns:
          tuple containing problem size, (N0, N1).
        """
        return self._N
            
    def get_preferences(self) :
        """ get solver preferences.

        Returns:
          dict: preference dictionary.

        References:
          `preference <preference.html>`_
        """
        prefs = {}
        prefs['tile_size'] = self._tile_size
        prefs['algorithm'] = self._get_algorithm()
        return prefs

    def get_optimize_dir(self) :
        """ get optimize direction
        
        Returns:
          optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
        """
        return self._optimize
    
    def get_E(self) :
        """ get QUBO energy.

        Returns:
          array of floating point number : QUBO energy.

        QUBO energy value is calculated for each trotter. so E is a vector whose length is m.

        Notes:
          You need to call calculate_E() or make_solution() before calling get_E().
          CPU/CUDA versions of solvers automatically call calculate_E() in get_E()
        """
        return self._E
        
    def get_x(self) :
        """ get bits.

        Returns:
          tuple of 2 numpy.int8 arrays : array of bit {0, 1}.

        x0.shape and x1.shape are (m, N0) and (m, N1) repsectively, and m is number of trotters.

        Note:
          calculate_E() or make_solution() should be called before calling get_E().
          ( CPU/CUDA annealers automatically/internally call calculate_E().)
        """
        return self._x

    def prepare(self) :
        """ preparation of internal resources.
        
        Note:
          prepare() should be called prior to run annealing loop.
        """
        
        N = self._N
        self._tile_size = min(1 << N, self._tile_size)
        self._xMax = 1 << N
        self._Emin = sys.float_info.max
        self._xbegin = 0

    def make_solution(self) :
        """ calculate QUBO energy.
        
        This method calculate QUBO energy, and caches it, does not return any value.

        Note:
          A call to this method can be asynchronous.
        """
        self.calculate_E()

    def calculate_E(self) :
        """ calculate QUBO energy.
        
        This method calculate QUBO energy, and caches it, does not return any value.

        Note:
          A call to this method can be asynchronous.
        """
        nMinX = len(self._x)
        self._E = np.empty((nMinX))
        self._E[...] = self._optimize.sign(self._Emin)
        
    def search_range(self) :
        """ Search minimum of QUBO energy within a range.

        Returns:
          True if search completed, False otherwise.

        This method enables stepwise minimum QUBO energy search.
        One method call covers a range specified by 'tile_size'.
        When a given problem is big, whole search takes very long time.
        By using this method, you can do searches stepwise.
        """
        
        xBegin = self._xbegin
        xEnd = self._tile_size
        
        N = self._N
        W = self._W
        xBegin = max(0, min(self._xMax, xBegin))
        xEnd = max(0, min(self._xMax, xEnd))
        x = sqaod.create_bitset_sequence(range(xBegin, xEnd), N)
        Etmp = formulas.dense_graph_batch_calculate_E(W, x)
        for i in range(xEnd - xBegin) :
            if self._Emin < Etmp[i] :
                continue
            elif Etmp[i] < self._Emin :
                self._Emin = Etmp[i]
                self._x = [x[i]]
            else :
                self._x.append(x[i])

        self._xbegin = xEnd
        return self._xbegin == self._xMax

    def search(self) :
        """ One liner for brute-force search. """
        self.prepare()
        xStep = min(self._tile_size, self._xMax)
        while not self.search_range() :
            pass
        self.make_solution()
        
    
def dense_graph_bf_searcher(W = None, optimize = sqaod.minimize, **prefs) :
    """ factory function for sqaod.py.DenseGraphAnnealer.

    Args:
      numpy.ndarray : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.py.DenseGraphBFSearcher: brute-force searcher instance
    """
    
    return DenseGraphBFSearcher(W, optimize, prefs)


if __name__ == '__main__' :

    W = np.array([[-32,4,4,4,4,4,4,4],
                  [4,-32,4,4,4,4,4,4],
                  [4,4,-32,4,4,4,4,4],
                  [4,4,4,-32,4,4,4,4],
                  [4,4,4,4,-32,4,4,4],
                  [4,4,4,4,4,-32,4,4],
                  [4,4,4,4,4,4,-32,4],
                  [4,4,4,4,4,4,4,-32]])
    
    
    N = 8
    bf = dense_graph_bf_searcher(W, sqaod.minimize)
    bf.search()
    E = bf.get_E()
    x = bf.get_x() 
    print(E, len(x), x[0])

    bf = dense_graph_bf_searcher(W, sqaod.maximize)
    bf.search()
    E = bf.get_E()
    x = bf.get_x() 
    print(E, len(x), x[0])



from __future__ import print_function
from __future__ import absolute_import
from .cuda_probe import is_cuda_available, cuda_failure_reason 
from ctypes.util import find_library
from ctypes import CDLL, c_int, c_char_p, POINTER, byref

    
def _check_shared_object(libname, version_getter) :
    try :
        so = CDLL(libname)
    except :
        return (False, libname, None)
    ver = version_getter(so)
    return (True, libname, ver)


def _sqaodc_version_getter(so) :
    funcname = 'sqaodc_version'
    if not hasattr(so, funcname) :
        return None
    func = getattr(so, funcname)
    func.restype = None
    func.argtype = [POINTER(c_int), POINTER(c_char_p)]
    ver = c_int(0)
    simd = c_char_p(0)
    func(byref(ver), byref(simd))
    return (ver.value, simd.value)

def check_libsqaodc() :
    return _check_shared_object('libsqaodc.so.1', _sqaodc_version_getter)

def _sqaodc_cuda_version_getter(so) :
    funcname = 'sqaodc_cuda_version'
    if not hasattr(so, funcname) :
        return None
    func = getattr(so, funcname)
    func.restype = None
    func.argtype = [POINTER(c_int), POINTER(c_int)]
    ver = c_int(0)
    cuda_ver = c_int(0)
    func(byref(ver), byref(cuda_ver))
    return (ver.value, cuda_ver.value)

def check_libsqaodc_cuda() :
    return _check_shared_object('libsqaodc_cuda.so.1', _sqaodc_cuda_version_getter)


def get_string_version(numver) :
    return '{}.{}.{}'.format(numver / 10000, (numver / 100) % 100, numver % 100)

def check_version(version, pkgver) :
    if version is None :
        return False
    strver = get_string_version(version[0])
    return strver == pkgver

def show_libraries() :
    libsqaodc = check_libsqaodc()
    libsqaodc_cuda = check_libsqaodc_cuda()
    print(libsqaodc)
    print(libsqaodc_cuda)


    

class EnvChecker :
    def __init__(self, pkgver) :
        self.pkgver = pkgver
    
    def check(self) :
        self.sqaodc = check_libsqaodc()
        self.sqaodc_cuda = check_libsqaodc_cuda()
        if not self.sqaodc[0] or not self.sqaodc_cuda[0] :
            return False
        return check_version(self.sqaodc[2], self.pkgver) and check_version(self.sqaodc_cuda[2], self.pkgver)
    
    def _format_message(self, lib) :
        success, libname, version = lib
        verok = False
        
        if not success :
            if libname is not None :
                err_reason = 'load failed'
            else :
                err_reason = 'library not found'
            verstr = 'N/A'
        else :
            verok = check_version(version, self.pkgver)
            verstr = get_string_version(version[0]) if verok else '< ' + str(self.pkgver)
            if not verok :
                err_reason = 'version mismatch'

        if success and verok :
            return 'OK {} ({})'.format(libname, verstr)
        else :
            return 'NG {} ({}) {}'.format(libname, verstr, err_reason)

    def show(self) :
        msg = self._format_message(self.sqaodc)
        print(msg)
        msg = self._format_message(self.sqaodc_cuda)
        print(msg)
    
        #print('CUDA                : {}'.format('Available' if cuda_available else 'not available'))
        #cuda_available = is_cuda_available()
    
if __name__ == '__main__' :
    show_libraries()
    checker = EnvChecker('1.0.0')
    checker.check()
    checker.show()



from __future__ import print_function
import sys
from sqaod.common import formulas_base as base
from . import device
from . import cuda_formulas as cext

# dense graph

def dense_graph_calculate_E(W, x, dtype) :
    return base.dense_graph_calculate_E(cext, _dgobj, W, x, dtype)

def dense_graph_batch_calculate_E(W, x, dtype) :
    return base.dense_graph_batch_calculate_E(cext, _dgobj, W, x, dtype)

def dense_graph_calculate_hamiltonian(W, dtype) :
    return base.dense_graph_calculate_hamiltonian(cext, _dgobj, W, dtype)

def dense_graph_calculate_E_from_spin(h, J, c, q, dtype) :
    return base.dense_graph_calculate_E_from_spin(cext, _dgobj, h, J, c, q, dtype)

def dense_graph_batch_calculate_E_from_spin(h, J, c, q, dtype) :
    return base.dense_graph_batch_calculate_E_from_spin(cext, _dgobj, h, J, c, q, dtype)


# bipartite_graph

def bipartite_graph_calculate_E(b0, b1, W, x0, x1, dtype) :
    return base.bipartite_graph_calculate_E(cext, _bgobj, b0, b1, W, x0, x1, dtype)

def bipartite_graph_batch_calculate_E(b0, b1, W, x0, x1, dtype) :
    return base.bipartite_graph_batch_calculate_E(cext, _bgobj, b0, b1, W, x0, x1, dtype)

def bipartite_graph_batch_calculate_E_2d(b0, b1, W, x0, x1, dtype) :
    return base.bipartite_graph_batch_calculate_E_2d(cext, _bgobj, b0, b1, W, x0, x1, dtype)

def bipartite_graph_calculate_hamiltonian(b0, b1, W, dtype) :
    return base.bipartite_graph_calculate_hamiltonian(cext, _bgobj, b0, b1, W, dtype)

def bipartite_graph_calculate_E_from_spin(h0, h1, J, c, q0, q1, dtype) :
    return base.bipartite_graph_calculate_E_from_spin(cext, _bgobj, h0, h1, J, c, q0, q1, dtype)

def bipartite_graph_batch_calculate_E_from_spin(h0, h1, J, c, q0, q1, dtype) :
    return base.bipartite_graph_batch_calculate_E_from_spin(cext, _bgobj, h0, h1, J, c, q0, q1, dtype)


# Global/Static

this_module = sys.modules[__name__]

def unload() :
    cext.dg_formulas_delete(this_module._dgobj)
    cext.bg_formulas_delete(this_module._bgobj)
    this_module._dgobj = None
    this_module._bgobj = None

if __name__ != "__main__" :
    this_module._dgobj = cext.dg_formulas_new()
    this_module._bgobj = cext.bg_formulas_new()
    # initialization
    cext.dg_formulas_assign_device(this_module._dgobj, device.active_device._cobj)
    cext.bg_formulas_assign_device(this_module._bgobj, device.active_device._cobj)
    import atexit
    atexit.register(unload)



from __future__ import print_function
import numpy as np
from sqaod.common.bipartite_graph_annealer_base import BipartiteGraphAnnealerBase
from . import cpu_bg_annealer as cext
import sqaod


class BipartiteGraphAnnealer(BipartiteGraphAnnealerBase) :
    
    def __init__(self, b0, b1, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        BipartiteGraphAnnealerBase.__init__(self, cext, dtype, b0, b1, W, optimize, prefdict)

        
def bipartite_graph_annealer(b0 = None, b1 = None, W = None, \
                             optimize = sqaod.minimize, \
                             dtype = np.float64, **prefs) :
    """ factory function for sqaod.cpu.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.cpu.BipartiteGraphAnnealer: annealer instance
    """
    return BipartiteGraphAnnealer(b0, b1, W, optimize, dtype, prefs)


if __name__ == '__main__' :
    N0 = 40
    N1 = 40
    m = 20
    
    np.random.seed(0)
            
    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5

    an = bipartite_graph_annealer(b0, b1, W, sqaod.minimize, n_trotters=m)
    
    Ginit = 5
    Gfin = 0.01
    beta = 1. / 0.02
    tau = 0.99
    n_repeat = 10

    for loop in range(0, n_repeat) :
        an.prepare()
        an.randomize_spin()
        G = Ginit
        while Gfin < G :
            an.anneal_one_step(G, beta)
            G = G * tau
        an.make_solution()

        E = an.get_E()
        x = an.get_x()
        print(E.shape, E)
        print(x)



import numpy as np
from . import common

# raising exceptions

def raise_not_a_scalar(caption, var) :
    tokens = [var, ' is not a scalar. ', str(type(var))]
    msg = ''.join(tokens)
    raise Exception(msg)

def raise_not_a_vector(caption, var) :
    tokens = [caption, ' is not a vector. ', str(type(var))]
    msg = ''.join(tokens)
    raise Exception(msg)

def raise_not_a_matrix(caption, var) :
    tokens = [caption, ' is not a matrix. ', str(type(var))]
    msg = ''.join(tokens)
    raise Exception(msg)

def raise_wrong_bit_type(catpion, var) :
    tokens = [var, ' is not the bit type(np.int8). ', str(type(var))]
    msg = ''.join(tokens)
    raise Exception(msg)

def raise_wrong_shape(caption, mat) :
    tokens = ['wrong shape dim=(', caption, ') = ']
    tokens.append(str(mat.shape))
    msg = ''.join(tokens)
    raise Exception(msg)

def raise_dims_dont_match(caption, varlist) :
    tokens = ['dimension mismatch for ', caption, '. ']
    for var in varlist :
        tokens.append(str(var.shape))
        tokens.append(', ')
    tokens.pop()
    msg = ''.join(tokens)
    raise Exception(msg)

# value/type checks

def assert_is_bits(bitslist) :
    for bits in bitslist :
        if bits.dtype != np.int8 :
            raise Exception("bits are not given as np.int8")

def assert_is_matrix(caption, var) :
    if not (hasattr(var, 'shape') and len(var.shape) == 2) :
        raise_not_a_matrix(caption, var)

def assert_is_vector(caption, vec) :
    if not common.is_vector(vec) :
        raise_not_a_vector(caption, vec)

def assert_is_scalar(caption, scalar) :
    if not common.is_scalar(scalar) :
        raise_not_a_scalar(caption, scalar)

def square_matrix(W, caption) :
    if not common.is_square(W) :
        raise_wrong_shape(caption, (W))
        
# dense graph    

class dense_graph :
    
    @staticmethod
    def qubo(W) :
        if len(W.shape) != 2 or W.shape[0] != W.shape[1] :
            raise_wrong_shape('W', W)

    # ising model dimension check    
    @staticmethod
    def hJc(h, J, c) :
        assert_is_vector('h', h)
        if len(J.shape) != 2 or J.shape[0] != J.shape[1] :
            raise_wrong_shape('J', J)
        assert_is_scalar('c', (c))
        
    @staticmethod
    def bits(W, x, caption) :
        assert_is_bits(x)
        matched = (W.shape[0] == x.shape[-1])
        if not matched :
            raise_dims_dont_match('W, x', (W, x))

    @staticmethod
    def x(W, x) :
        assert_is_bits(x)
        dense_graph.bits(W, x, 'x')

    @staticmethod
    def q(W, x) :
        assert_is_bits(x)
        dense_graph.bits(W, x, 'x')

    @staticmethod
    def xbatch(W, x) :
        assert_is_bits(x)
        dense_graph.bits(W, x, 'x')

    @staticmethod
    def qbatch(W, q) :
        assert_is_bits(q)
        dense_graph.bits(W, q, 'q')

            
# bipartite_graph

class bipartite_graph :
    @staticmethod
    def qubo(b0, b1, W) :
        assert_is_vector('b0', b0);
        assert_is_vector('b1', b1);
        if len(W.shape) != 2 :
            raise_wrong_shape('W', W)
        matched = (b0.shape[0] == W.shape[1]) and (b1.shape[0] == W.shape[0])
        if not matched :
            raise_dims_dont_match('b0, b1, W', (b0, b1, W))

    @staticmethod
    def hJc(h0, h1, J, c) :
        assert_is_vector('h0', h0);
        assert_is_vector('h1', h1);
        if len(J.shape) != 2 :
            raise_wrong_shape('J', J)
        assert_is_scalar('c', c)
        matched = (h0.shape[0] == J.shape[1]) and (h1.shape[0] == J.shape[0])
        if not matched :
            raise_dims_dont_match('h0, h1, J', (h0, h1, J))

    @staticmethod
    def bits(W, x0, x1, caption) :
        # dimension checks
        assert_is_bits(x0)
        assert_is_bits(x1)
        matched = (W.shape[1] == x0.shape[len(x0.shape) - 1]) and \
                  (W.shape[0] == x1.shape[len(x1.shape) - 1])
        if len(x0.shape) != 1 or len(x1.shape) != 1 :
            matched &= len(x0.shape) == 2 and len(x1.shape) == 2
            matched &= x0.shape[0] == x1.shape[0]
            
        if not matched :
            raise_dims_dont_match(caption, (x0, x1))

    @staticmethod
    def x(W, x0, x1) :
        assert_is_vector('x0', x0)
        assert_is_vector('x1', x1)
        bipartite_graph.bits(W, x0, x1, 'x0, x1')

    @staticmethod
    def q(W, q0, q1) :
        assert_is_vector('q0', q0)
        assert_is_vector('q1', q1)
        bipartite_graph.bits(W, q0, q1, 'q0, q1')

    @staticmethod
    def xbatch(W, x0, x1) :
        assert_is_matrix('x0', x0)
        assert_is_matrix('x1', x1)
        bipartite_graph.bits(W, x0, x1, 'x0, x1')

    @staticmethod
    def qbatch(W, q0, q1) :
        assert_is_matrix('q0', q0)
        assert_is_matrix('q1', q1)
        bipartite_graph.bits(W, q0, q1, 'q0, q1')



import numpy as np
from . import preference as pref
import inspect

def inherit(dst_cls, src_cls) :
    src_dict = src_cls.__dict__
    dst_dict = dst_cls.__dict__
    
    dst_keys =  dst_dict.keys()
    for name, attr in src_dict.items() :
        if inspect.isfunction(attr) and ((name in dst_keys) and hasattr(attr, '__doc__')) :
            attr.__doc__ = src_dict[name].__doc__
            
def copy(dst, src) :
    dst.__doc__ = src.__doc__



import numpy as np
import sys
import sqaod
from sqaod.common import checkers
from sqaod import algorithm as algo

class BipartiteGraphBFSearcher :
    """ Bipartite graph  brute-force searcher"""

    _tile_size_0_default = 512
    _tile_size_1_default = 512
    
    def __init__(self, W, b0, b1, optimize, prefdict) :
        self._verbose = False
        if not W is None :
            self.set_qubo(W, b0, b1, optimize)
        self._set_prefdict(prefdict)

    def _vars(self) :
        return self._b0, self._b1, self._W

    def get_problem_size(self) :
        """ get problem size.

        Problem size is defined as a number of bits of QUBO.

        Returns:
          tuple containing problem size, (N0, N1).
        """
        return self._N0, self._N1
    
    def set_qubo(self, b0, b1, W, optimize = sqaod.minimize) :
        """ set QUBO.

        Args:
          numpy.ndarray b0, b1, W : QUBO.
          optimize : optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
        checkers.bipartite_graph.qubo(b0, b1, W)
        """
        self._N0 = b0.shape[0]
        self._N1 = b1.shape[0]
        self._optimize = optimize
        self._b0 = self._optimize.sign(b0)
        self._b1 = self._optimize.sign(b1)
        self._W = self._optimize.sign(W)
        self._tile_size_0 = BipartiteGraphBFSearcher._tile_size_0_default
        self._tile_size_1 = BipartiteGraphBFSearcher._tile_size_1_default

    def _get_algorithm(self) :
        return algo.brute_force_search

    def set_preferences(self, prefdict=None, **prefs) :
        """ set solver preferences.

        Args:
          prefdict(dict) : preference dictionary.
          prefs(dict) : preference dictionary as \*\*kwargs. 

        References:
          `preference <preference.html>`_
        """
        if not prefdict is None :
            self._set_prefdict(prefdict)
        self._set_prefdict(prefs)
        
    def _set_prefdict(self, prefdict) :
        v = prefdict.get('tile_size_0')
        if v is not None :
            self._tile_size_0 = v;
        v = prefdict.get('tile_size_1')
        if v is not None :
            self._tile_size_1 = v;

    def get_preferences(self) :
        """ get solver preferences.

        Returns:
          dict: preference dictionary.

        References:
          `preference <preference.html>`_
        """
        prefs = {}
        prefs['tile_size_0'] = self._tile_size_0
        prefs['tile_size_1'] = self._tile_size_1
        prefs['algorithm'] = self._get_algorithm()
        return prefs

    def get_optimize_dir(self) :
        """ get optimize direction
        
        Returns:
          optimize direction, `sqaod.maximize, sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
        """
        return self._optimize
    
    def get_E(self) :
        """ get QUBO energy.

        Returns:
          array of floating point number : QUBO energy.

        QUBO energy value is calculated for each trotter. so E is a vector whose length is m.

        Notes:
          You need to call calculate_E() or make_solution() before calling get_E().
          CPU/CUDA versions of solvers automatically call calculate_E() in get_E()
        """
        return self._E

    def get_x(self) :
        """ get bits.

        Returns:
          tuple of 2 numpy.int8 arrays : array of bit {0, 1}.

        x0.shape and x1.shape are (m, N0) and (m, N1) repsectively, and m is number of trotters.

        Note:
          calculate_E() or make_solution() should be called before calling get_E().
          ( CPU/CUDA annealers automatically/internally call calculate_E().)
        """
        return self._xPairs

    def prepare(self) :
        """ preparation of internal resources.
        
        Note:
          prepare() should be called prior to run annealing loop.
        """
        N0, N1 = self.get_problem_size()
        self._x0begin = 0
        self._x1begin = 0
        self._x0max = 1 << N0
        self._x1max = 1 << N1
        self._tile_size_0 = min(self._x0max, self._tile_size_0)
        self._tile_size_1 = min(self._x1max, self._tile_size_1)
        self._Emin = sys.float_info.max
        self._xPairs = []

    def make_solution(self) :
        """ make bit arrays(x) and calculate QUBO energy.

        Note:
          A call to this method can be asynchronous.
        """
        self.calculate_E()

    def calculate_E(self) :
        """ calculate QUBO energy.
        
        This method calculate QUBO energy, and caches it, does not return any value.

        Note:
          A call to this method can be asynchronous.
        """
        nXmin = len(self._xPairs)
        self._E = np.empty((nXmin))
        self._E[...] = self._optimize.sign(self._Emin)

    # Not used.  Keeping it for reference.    
    def _search_naive(self) :
        N0, N1 = self._get_dim()
        b0, b1, W = self._vars()

        for i0 in range(self._x0max) :
            x0 = sqaod.create_bitset_sequence((i0), N0)
            for i1 in range(self._x1max) :
                x1 = sqaod.create_bitset_sequence((i1), N1)
                Etmp = np.dot(b0, x0.transpose()) + np.dot(b1, x1.transpose()) \
                       + np.dot(x1, np.matmul(W, x0.transpose()))
                if self._Emin < Etmp :
                    continue
                elif Etmp < self._Emin :
                    self._Emin = Etmp
                    self._xPairs = [(x0, x1)]
                else :
                    self._xPairs.append = [(x0, x1)]
        
    def search_range(self) :
        """ Search minimum of QUBO energy within a range.

        Returns:
          True if search completed, False otherwise.

        This method enables stepwise minimum QUBO energy search.
        One method call covers a range specified by 'tile_size'.
        When a given problem is big, whole search takes very long time.
        By using this method, you can do searches stepwise.
        """
        N0, N1 = self.get_problem_size()
        b0, b1, W = self._vars()
        x0begin = max(0, min(self._x0max, self._x0begin))
        x0end = max(0, min(self._x0max, x0begin + self._tile_size_0))
        x1begin = max(0, min(self._x1max, self._x1begin))
        x1end = max(0, min(self._x1max, x1begin + self._tile_size_1))

        x0step = x0end - x0begin
        x1step = x1end - x1begin

        bits0 = sqaod.create_bitset_sequence(range(x0begin, x0end), N0)
        bits1 = sqaod.create_bitset_sequence(range(x1begin, x1end), N1)
        Etmp = np.matmul(b0, bits0.T).reshape(1, x0step) \
               + np.matmul(b1, bits1.T).reshape(x1step, 1) + np.matmul(bits1, np.matmul(W, bits0.T))
        for x1 in range(x1step) :
            for x0 in range(x0step) :
                if self._Emin < Etmp[x1][x0] :
                    continue
                elif Etmp[x1][x0] < self._Emin :
                    self._Emin = Etmp[x1][x0]
                    self._xPairs = [(bits0[x0], bits1[x1])]
                else :
                    self._xPairs.append((bits0[x0], bits1[x1]))

        self._x1begin = x1end
        if self._x1begin != self._x1max :
            self._x1begin = x1end
            return False
        
        if self._x0begin == self._x0max :
            return True
        
        self._x1begin = 0
        self._x0begin = x0end
        return False

                    
    def search(self) :
        """ One liner for brute-force search. """
        self.prepare()

        while not self.search_range() :
            pass

        self.make_solution()
        
def bipartite_graph_bf_searcher(b0 = None, b1 = None, W = None, optimize = sqaod.minimize, **prefs) :
    """ factory function for sqaod.py.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.py.BipartiteGraphBFSearcher: annealer instance
    """
    return BipartiteGraphBFSearcher(b0, b1, W, optimize, prefs)


if __name__ == '__main__' :
    N0 = 14
    N1 = 9
    
    np.random.seed(0)

    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5
    
    bf = bipartite_graph_bf_searcher(b0, b1, W)
    bf.search()
    E = bf.get_E()
    x = bf.get_x() 
    print(E)
    print(x)
    print(bf.get_preferences())



import numpy as np


def sort_bitset_list(bitsSetList) :
    size = len(bitsSetList)
    nBitsSetList = len(bitsSetList[0])
    for bitsIdx in range(nBitsSetList) :
        N = bitsSetList[0][bitsIdx].shape[0]
        for pos in range(N - 1, -1, -1) :
            sets0, sets1 = [], []
            for idx in range(size) :
                if bitsSetList[idx][bitsIdx][pos] == 0 :
                    sets0.append(bitsSetList[idx])
                else :
                    sets1.append(bitsSetList[idx])
        bitsSetList = sets0 + sets1

    return bitsSetList


def sort_bitset(bitsSetList) :
    if type(bitsSetList[0]) is tuple or type(bitsSetList[0]) is list :
        return sort_bitset_list(bitsSetList)

    size = len(bitsSetList)
    nBitsSetList = len(bitsSetList)
    N = bitsSetList[0].shape[0]
    for pos in range(N - 1, -1, -1) :
        sets0, sets1 = [], []
        for idx in range(size) :
            if bitsSetList[idx][pos] == 0 :
                sets0.append(bitsSetList[idx])
            else :
                sets1.append(bitsSetList[idx])
        bitsSetList = sets0 + sets1
    return bitsSetList

def unique_x(x) :
    x = sort_bitset(x)
    u = []
    while len(x) != 0 :
        x0 = x[0]
        x.pop(0)
        u.append(x0)
        while len(x) != 0 and np.allclose(x0, x[0]) :
            x.pop(0)
            
    return u


class Summary :
    def __init__(self, solver = None) :
        if not solver is None :
            self.summarize(solver)

    def summarize(self, solver) :
        E = solver.get_E()
        xset = solver.get_x()

        optdir = solver.get_optimize_dir();
        self.E = optdir.best(E)
        bestX = []
        for idx in range(E.shape[0]) :
            if E[idx] == self.E :
                bestX.append(xset[idx])
                
        self.xlist = unique_x(bestX)

    
def make_summary(solver = None) :
    return Summary(solver)




from . import formulas

from .dense_graph_annealer import DenseGraphAnnealer
from .dense_graph_bf_searcher import DenseGraphBFSearcher
from .bipartite_graph_annealer import BipartiteGraphAnnealer
from .bipartite_graph_bf_searcher import BipartiteGraphBFSearcher

from .dense_graph_annealer import dense_graph_annealer
from .dense_graph_bf_searcher import dense_graph_bf_searcher
from .bipartite_graph_annealer import bipartite_graph_annealer
from .bipartite_graph_bf_searcher import bipartite_graph_bf_searcher




from __future__ import print_function
import numpy as np
import sqaod
from sqaod.common.dense_graph_bf_searcher_base import DenseGraphBFSearcherBase
from . import cpu_dg_bf_searcher as cext

class DenseGraphBFSearcher(DenseGraphBFSearcherBase) :

    def __init__(self, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        DenseGraphBFSearcherBase.__init__(self, cext, dtype, W, optimize, prefdict)


def dense_graph_bf_searcher(W = None, optimize = sqaod.minimize, dtype=np.float64, **prefs) :
    """ factory function for sqaod.cpu.DenseGraphAnnealer.

    Args:
      numpy.ndarray : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.cpu.DenseGraphBFSearcher: brute-force searcher instance
    """
    
    return DenseGraphBFSearcher(W, optimize, dtype, prefs)


if __name__ == '__main__' :

    np.random.seed(0)
    dtype = np.float32
    N = 8
    W = np.array([[-32,4,4,4,4,4,4,4],
                  [4,-32,4,4,4,4,4,4],
                  [4,4,-32,4,4,4,4,4],
                  [4,4,4,-32,4,4,4,4],
                  [4,4,4,4,-32,4,4,4],
                  [4,4,4,4,4,-32,4,4],
                  [4,4,4,4,4,4,-32,4],
                  [4,4,4,4,4,4,4,-32]])

    N = 16
    W = sqaod.generate_random_symmetric_W(N, -0.5, 0.5, dtype);
    bf = dense_graph_bf_searcher(W, sqaod.minimize, dtype)
    bf.set_preferences(tile_size=256)
    
    import time

    start = time.time()
    bf.search()
    end = time.time()
    print(end - start)
    
    x = bf.get_x() 
    E = bf.get_E()
    print(E)
    print(x)
    print(bf.get_preferences())
    
#    import sqaod.py.formulas
#    E = np.empty((1), dtype)
#    for bits in x :
#        print sqaod.py.formulas.dense_graph_calculate_E(W, bits)



from __future__ import print_function
import numpy as np
from . import checkers
from . import preference as pref
from . import common

class DenseGraphBFSearcherBase :
    
    def __init__(self, cext, dtype, W, optimize, prefdict) :
        self._cext = cext
        self.dtype = dtype
        if not W is None :
            self.set_qubo(W, optimize)
        self.set_preferences(prefdict)
            
    def __del__(self) :
        if hasattr(self, '_cobj') :
            self._cext.delete(self._cobj, self.dtype)

    def set_qubo(self, W, optimize = pref.minimize) :
        checkers.dense_graph.qubo(W)
        W = common.fix_type(W, self.dtype)
        W = common.symmetrize(W)
        self._N = W.shape[0]
        self._cext.set_qubo(self._cobj, W, optimize, self.dtype)
        self._optimize = optimize

    def get_problem_size(self) :
        return self._cext.get_problem_size(self._cobj, self.dtype)

    def set_preferences(self, prefdict = None, **prefs) :
        if not prefdict is None :
            self._cext.set_preferences(self._cobj, prefdict, self.dtype)
        self._cext.set_preferences(self._cobj, prefs, self.dtype)

    def get_preferences(self) :
        return self._cext.get_preferences(self._cobj, self.dtype);

    def get_optimize_dir(self) :
        return self._optimize

    def get_E(self) :
        return self._cext.get_E(self._cobj, self.dtype)

    def get_x(self) :
        return self._cext.get_x(self._cobj, self.dtype)

    def prepare(self) :
        self._cext.prepare(self._cobj, self.dtype);
        
    def make_solution(self) :
        self._cext.make_solution(self._cobj, self.dtype)

    def calculate_E(self) :
        self._cext.calculate_E(self._cobj, self.dtype)
        
    def search_range(self) :
        return self._cext.search_range(self._cobj, self.dtype)
        
    def search(self) :
        self.prepare()
        while True :
            comp, curx = self._cext.search_range(self._cobj, self.dtype)
            if comp :
                break;
        self.make_solution()
        
    def _search(self) :
        # one liner.  does not accept ctrl+c.
        self._cext.search(self._cobj, self.dtype)





from __future__ import print_function
import numpy as np
from . import checkers
from . import preference as pref
from . import common


class BipartiteGraphAnnealerBase :
    
    def __init__(self, cext, dtype, b0, b1, W, optimize, prefdict) :
        self._cext = cext
        self.dtype = dtype
        if not W is None :
            self.set_qubo(b0, b1, W, optimize)
        self.set_preferences(prefdict)

    def __del__(self) :
        if hasattr(self, '_cobj') :
            self._cext.delete(self._cobj, self.dtype)
        
    def seed(self, seed) :
        self._cext.seed(self._cobj, seed, self.dtype)
            
    def set_qubo(self, b0, b1, W, optimize = pref.minimize) :
        checkers.bipartite_graph.qubo(b0, b1, W)
        b0, b1, W = common.fix_type([b0, b1, W], self.dtype)
        self._cext.set_qubo(self._cobj, b0, b1, W, optimize, self.dtype);
        self._optimize = optimize

    def get_optimize_dir(self) :
        return self._optimize

    def get_problem_size(self) :
        return self._cext.get_problem_size(self._cobj, self.dtype)

    def set_preferences(self, prefdict=None, **prefs) :
        if not prefdict is None :
            self._cext.set_preferences(self._cobj, prefdict, self.dtype)
        self._cext.set_preferences(self._cobj, prefs, self.dtype)

    def get_preferences(self) :
        return self._cext.get_preferences(self._cobj, self.dtype)
        
    def get_E(self) :
        return self._cext.get_E(self._cobj, self.dtype)

    def get_x(self) :
        return self._cext.get_x(self._cobj, self.dtype)

    def set_q(self, qpair) :
        q0, q1 = qpair[0], qpair[1]
        if q0.dtype != np.int8 :
            q0 = np.asarray(q0, np.int8)
        if q1.dtype != np.int8 :
            q1 = np.asarray(q1, np.int8)
        self._cext.set_q(self._cobj, (q0, q1), self.dtype)

    def set_qset(self, qpairset) :
        qin = []
        for qpair in qpairset :
            q0, q1 = qpair[0], qpair[1]
            if q0.dtype != np.int8 :
                q0 = np.asarray(q0, np.int8)
            if q1.dtype != np.int8 :
                q1 = np.asarray(q1, np.int8)
            qin.append((q0, q1))
        self._cext.set_qset(self._cobj, qin, self.dtype)

    # Ising model / spins
    
    def get_hamiltonian(self) :
        N0, N1 = self.get_problem_size()
        h0 = np.ndarray((N0), self.dtype);
        h1 = np.ndarray((N1), self.dtype);
        J = np.ndarray((N1, N0), self.dtype);
        c = np.ndarray((1), self.dtype)
        self._cext.get_hamiltonian(self._cobj, h0, h1, J, c, self.dtype)
        return h0, h1, J, c[0]
    
    def set_hamiltonian(self, h0, h1, J, c) :
        checkers.bipartite_graph.hJc(h0, h1, J, c)
        h0, h1, J = common.fix_type([h0, h1, J], self.dtype)
        c = self.dtype(c)
        self._cext.set_hamiltonian(self._cobj, h0, h1, J, c, self.dtype)
        self._optimize = pref.minimize
            
    def get_q(self) :
        return self._cext.get_q(self._cobj, self.dtype)

    def randomize_spin(self) :
        self._cext.randomize_spin(self._cobj, self.dtype)

    def calculate_E(self) :
        self._cext.calculate_E(self._cobj, self.dtype)

    def prepare(self) :
        self._cext.prepare(self._cobj, self.dtype)

    def make_solution(self) :
        self._cext.make_solution(self._cobj, self.dtype)
        
    def anneal_one_step(self, G, beta) :
        G, beta = self.dtype(G), self.dtype(beta)
        self._cext.anneal_one_step(self._cobj, G, beta, self.dtype)




class Solver :
    
    @abstractmethod
    def select_algorithm(self, algo) :
        raise NotImplementedError()

    @abstractmethod
    def get_preferences(self, **kwargs) :
        raise NotImplementedError()

    @abstractmethod
    def set_preferences(self, **kwargs) :
        raise NotImplementedError()

    @abstractmethod
    def get_E(self) :
        raise NotImplementedError()

class BFSearcher(Solver) :

    @abstractmethod
    def prepare(self) :
        raise NotImplementedError()

    @abstractmethod
    def make_solution(self) :
        raise NotImplementedError()

    @abstractmethod
    def search(self) :
        raise NotImplementedError()

    
class Annealer(Sovler) :
    
    @abstractmethod
    def seed(self, seed) :
        raise NotImplementedError()

    @abstractmethod
    def randomize_spin(self) :
        raise NotImplementedError()

    @abstractmethod
    def prepare(self) :
        raise NotImplementedError()

    @abstractmethod
    def make_solution(self) :
        raise NotImplementedError()

    @abstractmethod
    def calculate_E(self) :
        raise NotImplementedError()

    @abstractmethod
    def anneal_one_step(self, G, beta) :
        raise NotImplementedError()

    
class DenseGraphSolver :
    @abstractmethod
    def get_problem_size(self) : # returns N.
        raise NotImplementedError()

    @abstractmethod
    def set_qubo(self, W, optimize = sqaod.minimize) :
        raise NotImplementedError()

    @abstractmethod
    def get_x(self) : # returns x.
        raise NotImplementedError()

    
class BipartiteGraphSolver :
    @abstractmethod
    def get_problem_size(self) : # returns N0, N1.
        raise NotImplementedError()

    @abstractmethod
    def set_qubo(self, b0, b1, W, optimize = sqaod.minimize) :
        raise NotImplementedError()

    @abstractmethod
    def get_x(self) : # returns x0 and x1.
        raise NotImplementedError()
                             

class DenseGraphBFSearcher(BFSearcher, DenseGraphSolver) :
    @abstractmethod
    def search_range(self, xBegin, xEnd) :
        raise NotImplementedError()

    def search(self) : # to be implemented.
        raise NotImplementedError()
    
    
class DenseGraphAnnealer(Annealer, DenseGraphSolver) :
    @abstractmethod
    def set_hamiltonian(self, h, J, c = 0.) :
        raise NotImplementedError()

    @abstractmethod
    def get_hamiltonian(self, x) :
        raise NotImplementedError()

    @abstractmethod
    def set_q(self, x) :
        raise NotImplementedError()

    @abstractmethod
    def get_q(self, x) :
        raise NotImplementedError()

    
class BipartiteGraphBFSearcher(BFSearcher, BipartiteGraphSolver) :
    @abstractmethod
    def search_range(self, x0Begin, x0End, x1Begin, x1End) :
        raise NotImplementedError()

    def search(self) : # to be implemented.
        raise NotImplementedError()
    
class BipartiteGraphAnnealer(Annealer, BipartiteGraphSolver) :
    @abstractmethod
    def set_hamiltonian(self, h0, h1, J, c = 0.) :
        raise NotImplementedError()

    @abstractmethod
    def get_hamiltonian(self, x0, x1) :
        raise NotImplementedError()

    @abstractmethod
    def set_q(self, x0, x1) :
        raise NotImplementedError()

    @abstractmethod
    def get_q(self, x) :
        raise NotImplementedError()




from setuptools import setup, find_packages, Extension, dist
from sysconfig import get_config_vars
import numpy as np
import sys
import os
from ctypes.util import find_library

name = 'sqaod'
version = '1.0.1'

npinclude = np.get_include()

(opt,) = get_config_vars('OPT')
os.environ['OPT'] = " ".join(
    flag for flag in opt.split() if flag != '-Wstrict-prototypes'
)


def extension(pkg, name) :
    ext = Extension('sqaod/{}/{}'.format(pkg, name),
                    include_dirs = [npinclude],
                    libraries = ['sqaodc'],
                    library_dirs = ['/usr/lib', 'sqaod/{}'.format(pkg)],
                    sources = ['sqaod/{}/src/{}.cpp'.format(pkg, name)],
                    extra_compile_args = ['-std=c++11', '-Wno-format-security'])
    return ext

def cuda_extension(pkg, name) :
    ext = Extension('sqaod/{}/{}'.format(pkg, name),
                    include_dirs = [npinclude],
                    libraries = ['sqaodc_cuda', 'sqaodc'],
                    library_dirs = ['/usr/lib', 'sqaod/{}'.format(pkg)],
                    sources = ['sqaod/{}/src/{}.cpp'.format(pkg, name)],
                    extra_compile_args = ['-std=c++11', '-Wno-format-security'])
    return ext

ext_modules = []
ext_modules.append(extension('cpu', 'cpu_formulas'))
ext_modules.append(extension('cpu', 'cpu_dg_bf_searcher'))
ext_modules.append(extension('cpu', 'cpu_dg_annealer'))
ext_modules.append(extension('cpu', 'cpu_bg_bf_searcher'))
ext_modules.append(extension('cpu', 'cpu_bg_annealer'))
if find_library('sqaodc_cuda') is not None :
    ext_modules.append(cuda_extension('cuda', 'cuda_device'))
    ext_modules.append(cuda_extension('cuda', 'cuda_formulas'))
    ext_modules.append(cuda_extension('cuda', 'cuda_dg_bf_searcher'))
    ext_modules.append(cuda_extension('cuda', 'cuda_dg_annealer'))
    ext_modules.append(cuda_extension('cuda', 'cuda_bg_bf_searcher'))
    ext_modules.append(cuda_extension('cuda', 'cuda_bg_annealer'))

pyver= [
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
]

email = 'shin.morino@gmail.com'    
author='Shinya Morino'


classifiers=[
    'Operating System :: POSIX :: Linux',
    'Natural Language :: English',
    'License :: OSI Approved :: Apache Software License',
    'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    'License :: OSI Approved :: BSD License',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Information Analysis'
]

classifiers = classifiers + pyver
url = 'https://github.com/shinmorino/sqaod/'

with open('README.rst') as file:
    long_description = file.read()

setup(
    name=name,
    version=version,
    url=url,
    author=author,
    author_email=email,
    maintainer=author,
    maintainer_email=email,
    description='A collection of solvers for simulated quantum annealing.',
    long_description=long_description,
    packages=find_packages(exclude=['tests']),
    install_requires=['numpy>=1.11'],
    keywords='Simulated quantum annealing, Quantum annealing, Quantum computing, Monte Carlo, OpenMP, GPU, CUDA',
    classifiers=classifiers,
    ext_modules=ext_modules
)



from __future__ import print_function
import numpy as np
import sqaod
from sqaod.common.bipartite_graph_bf_searcher_base import BipartiteGraphBFSearcherBase
from . import cpu_bg_bf_searcher as cext


class BipartiteGraphBFSearcher(BipartiteGraphBFSearcherBase) :
    def __init__(self, b0, b1, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        BipartiteGraphBFSearcherBase.__init__(self, cext, dtype, b0, b1, W, optimize, prefdict)
        

def bipartite_graph_bf_searcher(b0 = None, b1 = None, W = None, optimize = sqaod.minimize, dtype = np.float64, **prefs) :
    """ factory function for sqaod.cpu.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.cpu.BipartiteGraphBFSearcher: annealer instance
    """
    return BipartiteGraphBFSearcher(b0, b1, W, optimize, dtype, prefs)


if __name__ == '__main__' :
    N0 = 14
    N1 = 5
    
    np.random.seed(0)

    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5
    
    bf = bipartite_graph_bf_searcher(b0, b1, W)
    bf.set_preferences(tile_size_0=256, tile_size_1=256)
    bf.search()
    E = bf.get_E()
    x = bf.get_x() 
    print(E.shape, E)
    print(x)
    print(bf.get_preferences())



from __future__ import print_function
import numpy as np
from . import checkers
from . import preference as pref
from . import common

class DenseGraphAnnealerBase :
    
    def __init__(self, cext, dtype, W, optimize, prefdict) :
        self.dtype = dtype
        self._cext = cext
        if not W is None :
            self.set_qubo(W, optimize)
        self.set_preferences(prefdict)

    def __del__(self) :
        if hasattr(self, '_cobj') :
            self._cext.delete(self._cobj, self.dtype)
        
    def seed(self, seed) :
        self._cext.seed(self._cobj, seed, self.dtype)
        
    def set_qubo(self, W, optimize = pref.minimize) :
        checkers.dense_graph.qubo(W)
        W = common.fix_type(W, self.dtype)
        W = common.symmetrize(W)
        
        self._cext.set_qubo(self._cobj, W, optimize, self.dtype)
        self._optimize = optimize

    def get_problem_size(self) :
        return self._cext.get_problem_size(self._cobj, self.dtype)

    def set_preferences(self, prefdict = None, **prefs) :
        if not prefdict is None:
            self._cext.set_preferences(self._cobj, prefdict, self.dtype)
        self._cext.set_preferences(self._cobj, prefs, self.dtype)

    def get_preferences(self) :
        return self._cext.get_preferences(self._cobj, self.dtype)

    def get_optimize_dir(self) :
        return self._optimize

    def get_E(self) :
        return self._cext.get_E(self._cobj, self.dtype)

    def calculate_E(self) :
        self._cext.calculate_E(self._cobj, self.dtype)

    def get_x(self) :
        return self._cext.get_x(self._cobj, self.dtype)

    def set_q(self, q) :
        q = common.fix_type(q, np.int8)
        self._cext.set_q(self._cobj, q, self.dtype)

    def set_qset(self, qset) :
        qlist = []
        for qvec in qset :
            qvec = common.fix_type(qvec, np.int8)
            qlist.append(qvec)
        self._cext.set_qset(self._cobj, qlist, self.dtype)

    def get_hamiltonian(self) :
        N = self.get_problem_size()
        h = np.empty((N), self.dtype)
        J = np.empty((N, N), self.dtype)
        c = np.empty((1), self.dtype)
        self._cext.get_hamiltonian(self._cobj, h, J, c, self.dtype)
        return h, J, c[0]

    def set_hamiltonian(self, h, J, c) :
        checkers.dense_graph.hJc(h, J, c)
        
        h, J = common.fix_type([h, J], self.dtype)
        J = common.symmetrize(J)
        
        c = self.dtype(c)
        self._cext.set_hamiltonian(self._cobj, h, J, c, self.dtype)
        self._optimize = pref.minimize

    def get_q(self) :
        return self._cext.get_q(self._cobj, self.dtype)

    def randomize_spin(self) :
        self._cext.randomize_spin(self._cobj, self.dtype)

    def prepare(self) :
        self._cext.prepare(self._cobj, self.dtype)

    def make_solution(self) :
        self._cext.make_solution(self._cobj, self.dtype)

    def anneal_one_step(self, G, beta) :
        G, beta = self.dtype(G), self.dtype(beta)
        self._cext.anneal_one_step(self._cobj, G, beta, self.dtype)



from __future__ import print_function
import numpy as np
from . import checkers
from . import preference as pref
from . import common

class BipartiteGraphBFSearcherBase :
    
    def __init__(self, cext, dtype, b0, b1, W, optimize, prefdict) :
        self._cext = cext
        self.dtype = dtype
        self._optimize = optimize
        if not W is None :
            self.set_qubo(b0, b1, W, optimize)
        self.set_preferences(prefdict)
            
    def __del__(self) :
        self._cext.delete(self._cobj, self.dtype)

    def set_qubo(self, b0, b1, W, optimize = pref.minimize) :
        checkers.bipartite_graph.qubo(b0, b1, W)
        b0, b1, W = common.fix_type([b0, b1, W], self.dtype)
        self._dim = (b0.shape[0], b1.shape[0])
        self._cext.set_qubo(self._cobj, b0, b1, W, optimize, self.dtype)

    def get_problem_size(self) :
        return self._cext.get_problem_size(self._cobj, self.dtype)

    def set_preferences(self, prefdict=None, **prefs) :
        if not prefdict is None :
            self._cext.set_preferences(self._cobj, prefdict, self.dtype)
        self._cext.set_preferences(self._cobj, prefs, self.dtype)

    def get_preferences(self) :
        return self._cext.get_preferences(self._cobj, self.dtype);
        
    def get_optimize_dir(self) :
        return self._optimize

    def get_E(self) :
        return self._cext.get_E(self._cobj, self.dtype);

    def get_x(self) :
        return self._cext.get_x(self._cobj, self.dtype);
    
    def prepare(self) :
        self._cext.prepare(self._cobj, self.dtype);
        
    def make_solution(self) :
        self._cext.make_solution(self._cobj, self.dtype);
        
    def calculate_E(self) :
        self._cext.calculate_E(self._cobj, self.dtype);
        
    def search_range(self) :
        return self._cext.search_range(self._cobj, self.dtype)
        
    def _search(self) :
        # one liner.  not interruptive.
        self._cext.search(self._cobj, self.dtype);

    def search(self) :
        self.prepare()
        # users is able to use this search loop on his code to get intermediate states
        # such as progress and solutions.
        while True :
            comp, curx0, curx1 = self._cext.search_range(self._cobj, self.dtype)
            if comp :
                break;
        self.make_solution()



from __future__ import print_function
import numpy as np
import sqaod
from sqaod.common.bipartite_graph_annealer_base import BipartiteGraphAnnealerBase
from sqaod.common import docstring
from . import cuda_bg_annealer as cext
from . import device


class BipartiteGraphAnnealer(BipartiteGraphAnnealerBase) :

    def __init__(self, b0, b1, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        cext.assign_device(self._cobj, device.active_device._cobj, dtype)
        self._device = device.active_device
        BipartiteGraphAnnealerBase.__init__(self, cext, dtype, b0, b1, W, optimize, prefdict)


def bipartite_graph_annealer(b0 = None, b1 = None, W = None, \
                             optimize = sqaod.minimize, dtype = np.float64, **prefs) :
    """ factory function for sqaod.cuda.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.cuda.BipartiteGraphAnnealer: annealer instance
    """
    return BipartiteGraphAnnealer(b0, b1, W, optimize, dtype, prefs)

# inherit docstring from interface
docstring.inherit(BipartiteGraphAnnealer, sqaod.py.BipartiteGraphAnnealer)

if __name__ == '__main__' :
    N0 = 40
    N1 = 40
    m = 20
    
    np.random.seed(0)
            
    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5

    an = bipartite_graph_annealer(b0, b1, W, sqaod.minimize, n_trotters=m)
    
    Ginit = 5
    Gfin = 0.01
    beta = 1. / 0.02
    tau = 0.99
    n_repeat = 10

    for loop in range(0, n_repeat) :
        an.prepare()
        an.randomize_spin()
        G = Ginit
        while Gfin < G :
            an.anneal_one_step(G, beta)
            G = G * tau
        an.make_solution()

        E = an.get_E()
        x = an.get_x()
        print(E.shape, E)
        print(x)



from __future__ import print_function
import numpy as np
import sqaod
from sqaod.common.bipartite_graph_bf_searcher_base import BipartiteGraphBFSearcherBase
from sqaod.common import docstring
from . import cuda_bg_bf_searcher as cext
from . import device

class BipartiteGraphBFSearcher(BipartiteGraphBFSearcherBase) :
    
    def __init__(self, b0, b1, W, optimize, dtype, prefdict) :
        self._cobj = cext.new(dtype)
        cext.assign_device(self._cobj, device.active_device._cobj, dtype)
        self._device = device.active_device
        BipartiteGraphBFSearcherBase.__init__(self, cext, dtype, b0, b1, W, optimize, prefdict)


def bipartite_graph_bf_searcher(b0 = None, b1 = None, W = None,
                                optimize = sqaod.minimize, dtype = np.float64,
                                **prefs) :
    """ factory function for sqaod.cuda.BipartiteGraphAnnealer.

    Args:
      numpy.ndarray b0, b1, W : QUBO
      optimize : specify optimize direction, `sqaod.maximize or sqaod.minimize <preference.html#sqaod-maximize-sqaod-minimize>`_.
      prefs : `preference <preference.html>`_ as \*\*kwargs
    Returns:
      sqaod.cuda.BipartiteGraphBFSearcher: annealer instance
    """
    searcher = BipartiteGraphBFSearcher(b0, b1, W, optimize, dtype, prefs)
    return searcher

# inherit docstring from interface
docstring.inherit(BipartiteGraphBFSearcher, sqaod.py.BipartiteGraphBFSearcher)

if __name__ == '__main__' :
    N0 = 14
    N1 = 5
    
    np.random.seed(0)

    W = np.random.random((N1, N0)) - 0.5
    b0 = np.random.random((N0)) - 0.5
    b1 = np.random.random((N1)) - 0.5
    
    bf = bipartite_graph_bf_searcher(b0, b1, W)
    bf.search()
    E = bf.get_E()
    x = bf.get_x() 
    print(E.shape, E)
    print(x)
